@@ -25,8 +25,8 @@
 extern "C" {
 #endif
 
-#include <opensc/opensc.h>
-#include <opensc/pkcs15.h>
+#include "libopensc/opensc.h"
+#include "libopensc/pkcs15.h"
 
 struct sc_asn1_entry {
 	const char *name;
@@ -99,6 +99,8 @@ int sc_asn1_decode_bit_string_ni(const u8 * inbuf, size_t inlen,
 int sc_asn1_decode_integer(const u8 * inbuf, size_t inlen, int *out);
 int sc_asn1_decode_object_id(const u8 * inbuf, size_t inlen,
 			     struct sc_object_id *id);
+int sc_asn1_encode_object_id(u8 **buf, size_t *buflen, 
+				const struct sc_object_id *id);
 
 /* algorithm encoding/decoding */
 int sc_asn1_decode_algorithm_id(struct sc_context *,
@@ -122,6 +124,7 @@ void sc_asn1_clear_algorithm_id(struct sc_algorithm_id *);
 #define SC_ASN1_OPTIONAL		0x00000002
 #define SC_ASN1_ALLOC			0x00000004
 #define SC_ASN1_UNSIGNED		0x00000008
+#define SC_ASN1_EMPTY_ALLOWED           0x00000010
 
 #define SC_ASN1_BOOLEAN                 1
 #define SC_ASN1_INTEGER                 2
