@@ -18,13 +18,16 @@
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
-#include "internal.h"
-#include "asn1.h"
-#include "cardctl.h"
+#include "config.h"
+
 #include <stdlib.h>
 #include <string.h>
 #include <ctype.h>
 
+#include "internal.h"
+#include "asn1.h"
+#include "cardctl.h"
+
 static struct sc_atr_table pgp_atrs[] = {
 	{ "3b:fa:13:00:ff:81:31:80:45:00:31:c1:73:c0:01:00:00:90:00:b1", NULL, NULL, SC_CARD_TYPE_OPENPGP_GENERIC, 0, NULL },
 	{ NULL, NULL, NULL, 0, 0, NULL }
@@ -127,7 +130,7 @@ pgp_init(sc_card_t *card)
 	struct do_info	*info;
 	int		r;
 
-	priv = (struct pgp_priv_data *) calloc (1, sizeof *priv);
+	priv = calloc (1, sizeof *priv);
 	if (!priv)
 		return SC_ERROR_OUT_OF_MEMORY;
 	card->name = "OpenPGP";
@@ -192,7 +195,7 @@ pgp_set_blob(struct blob *blob, const u8 *data, size_t len)
 		free(blob->data);
 	blob->len    = len;
 	blob->status = 0;
-	blob->data   = (unsigned char *) malloc(len);
+	blob->data   = malloc(len);
 	memcpy(blob->data, data, len);
 
 	blob->file->size = len;
@@ -206,7 +209,7 @@ pgp_new_blob(struct blob *parent, unsigned int file_id,
 	sc_file_t	*file = sc_file_new();
 	struct blob	*blob, **p;
 
-	blob = (struct blob *) calloc(1, sizeof(*blob));
+	blob = calloc(1, sizeof(*blob));
 	blob->parent = parent;
 	blob->id     = file_id;
 	blob->file   = file;
@@ -235,9 +238,7 @@ pgp_read_blob(sc_card_t *card, struct blob *blob)
 	if (blob->info == NULL)
 		return blob->status;
 
-	sc_ctx_suppress_errors_on(card->ctx);
 	r = blob->info->get_fn(card, blob->id, buffer, sizeof(buffer));
-	sc_ctx_suppress_errors_off(card->ctx);
 
 	if (r < 0) {
 		blob->status = r;
@@ -308,7 +309,7 @@ pgp_enumerate_blob(sc_card_t *card, struct blob *blob)
 
 	return 0;
 
-eoc:	sc_error(card->ctx, "Unexpected end of contents\n");
+eoc:	sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "Unexpected end of contents\n");
 	return SC_ERROR_OBJECT_NOT_VALID;
 }
 
@@ -441,7 +442,7 @@ pgp_get_pubkey(sc_card_t *card, unsigned int tag, u8 *buf, size_t buf_len)
 	u8		idbuf[2];
 	int		r;
 
-	sc_debug(card->ctx, "called, tag=%04x\n", tag);
+	sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "called, tag=%04x\n", tag);
 
 	idbuf[0] = tag >> 8;
 	idbuf[1] = tag;
@@ -455,9 +456,9 @@ pgp_get_pubkey(sc_card_t *card, unsigned int tag, u8 *buf, size_t buf_len)
 	apdu.resplen = buf_len;
 
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, r, "APDU transmit failed");
+	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
 	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
-	SC_TEST_RET(card->ctx, r, "Card returned error");
+	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "Card returned error");
 
 	return apdu.resplen;
 }
@@ -472,7 +473,7 @@ pgp_get_pubkey_pem(sc_card_t *card, unsigned int tag, u8 *buf, size_t buf_len)
 	size_t		len;
 	int		r;
 
-	sc_debug(card->ctx, "called, tag=%04x\n", tag);
+	sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "called, tag=%04x\n", tag);
 	
 	if ((r = pgp_get_blob(card, &priv->mf, tag & 0xFFFE, &blob)) < 0
 	 || (r = pgp_get_blob(card, blob, 0x7F49, &blob)) < 0
@@ -512,9 +513,9 @@ pgp_get_data(sc_card_t *card, unsigned int tag, u8 *buf, size_t buf_len)
 	apdu.resplen = buf_len;
 
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, r, "APDU transmit failed");
+	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
 	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
-	SC_TEST_RET(card->ctx, r, "Card returned error");
+	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "Card returned error");
 
 	return apdu.resplen;
 }
@@ -556,7 +557,7 @@ pgp_set_security_env(sc_card_t *card,
 	case SC_SEC_OPERATION_SIGN:
 		if (env->key_ref[0] != 0x00
 		 && env->key_ref[0] != 0x02) {
-		 	sc_error(card->ctx,
+		 	sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
 				"Key reference not compatible with "
 				"requested usage\n");
 			return SC_ERROR_NOT_SUPPORTED;
@@ -564,7 +565,7 @@ pgp_set_security_env(sc_card_t *card,
 		break;
 	case SC_SEC_OPERATION_DECIPHER:
 		if (env->key_ref[0] != 0x01) {
-		 	sc_error(card->ctx,
+		 	sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
 				"Key reference not compatible with "
 				"requested usage\n");
 			return SC_ERROR_NOT_SUPPORTED;
@@ -602,11 +603,11 @@ pgp_compute_signature(sc_card_t *card, const u8 *data,
 				0x88, 0, 0);
 		break;
 	case 0x01:
-		sc_error(card->ctx,
+		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
 			"Invalid key reference (decipher only key)\n");
 		return SC_ERROR_INVALID_ARGUMENTS;
 	default:
-		sc_error(card->ctx, "Invalid key reference 0x%02x\n",
+		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "Invalid key reference 0x%02x\n",
 				env->key_ref[0]);
 		return SC_ERROR_INVALID_ARGUMENTS;
 	}
@@ -619,9 +620,9 @@ pgp_compute_signature(sc_card_t *card, const u8 *data,
 	apdu.resplen = outlen;
 
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, r, "APDU transmit failed");
+	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
 	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
-	SC_TEST_RET(card->ctx, r, "Card returned error");
+	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "Card returned error");
 
 	return apdu.resplen;
 }
@@ -638,7 +639,7 @@ pgp_decipher(sc_card_t *card, const u8 *in, size_t inlen,
 
 	/* There's some funny padding indicator that must be
 	 * prepended... hmm. */
-	if (!(temp = (u8 *) malloc(inlen + 1)))
+	if (!(temp = malloc(inlen + 1)))
 		return SC_ERROR_OUT_OF_MEMORY;
 	temp[0] = '\0';
 	memcpy(temp + 1, in, inlen);
@@ -658,12 +659,12 @@ pgp_decipher(sc_card_t *card, const u8 *in, size_t inlen,
 		break;
 	case 0x00: /* signature key */
 	case 0x02: /* authentication key */
-		sc_error(card->ctx,
+		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
 			"Invalid key reference (signature only key)\n");
 		free(temp);
 		return SC_ERROR_INVALID_ARGUMENTS;
 	default:
-		sc_error(card->ctx, "Invalid key reference 0x%02x\n",
+		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "Invalid key reference 0x%02x\n",
 				env->key_ref[0]);
 		free(temp);
 		return SC_ERROR_INVALID_ARGUMENTS;
@@ -679,9 +680,9 @@ pgp_decipher(sc_card_t *card, const u8 *in, size_t inlen,
 	r = sc_transmit_apdu(card, &apdu);
 	free(temp);
 
-	SC_TEST_RET(card->ctx, r, "APDU transmit failed");
+	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
 	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
-	SC_TEST_RET(card->ctx, r, "Card returned error");
+	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "Card returned error");
 
 	return apdu.resplen;
 }
