@@ -20,12 +20,14 @@
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
-#include "internal.h"
-#include "cardctl.h"
+#include "config.h"
+
 #include <stdlib.h>
 #include <string.h>
 
-#include <opensc/asn1.h>
+#include "internal.h"
+#include "asn1.h"
+#include "cardctl.h"
 
 static struct sc_atr_table setcos_atrs[] = {
 	/* some Nokia branded SC */
@@ -70,15 +72,10 @@ static struct sc_card_driver setcos_drv = {
 	NULL, 0, NULL
 };
 
-static int setcos_finish(sc_card_t *card)
-{
-	return 0;
-}
-
 static int match_hist_bytes(sc_card_t *card, const char *str, size_t len)
 {
-	const char *src = (const char *) card->slot->atr_info.hist_bytes;
-	size_t srclen = card->slot->atr_info.hist_bytes_len;
+	const char *src = (const char *) card->reader->atr_info.hist_bytes;
+	size_t srclen = card->reader->atr_info.hist_bytes_len;
 	size_t offset = 0;
 
 	if (len == 0)
@@ -125,7 +122,7 @@ static int setcos_match_card(sc_card_t *card)
 				card->type = SC_CARD_TYPE_SETCOS_EID_V2_1;
 			else {
 				buf[sizeof(buf) - 1] = '\0';
-				sc_debug(card->ctx, "SetCOS EID applet %s is not supported", (char *) buf);
+				sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "SetCOS EID applet %s is not supported", (char *) buf);
 				return 0;
 			}
 			return 1;
@@ -145,9 +142,7 @@ static int select_pkcs15_app(sc_card_t * card)
 	/* Regular PKCS#15 AID */
 	sc_format_path("A000000063504B43532D3135", &app);
 	app.type = SC_PATH_TYPE_DF_NAME;
-	sc_ctx_suppress_errors_on(card->ctx);
 	r = sc_select_file(card, &app, NULL);
-	sc_ctx_suppress_errors_off(card->ctx);
 	return r;
 }
 
@@ -474,7 +469,7 @@ static int setcos_create_file_44(sc_card_t *card, sc_file_t *file)
 				break;
 			case SC_AC_CHV:				/* pin */
 				if ((bNumber & 0x7F) == 0 || (bNumber & 0x7F) > 7) {
-					sc_error(card->ctx, "SetCOS 4.4 PIN refs can only be 1..7\n");
+					sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "SetCOS 4.4 PIN refs can only be 1..7\n");
 					return SC_ERROR_INVALID_ARGUMENTS;
 				}
 				bCommands_pin[setcos_pin_index_44(pins, sizeof(pins), (int) bNumber)] |= 1 << i;
@@ -577,11 +572,11 @@ static int setcos_set_security_env2(sc_card_t *card,
 	    card->type == SC_CARD_TYPE_SETCOS_NIDEL ||
 	    SETCOS_IS_EID_APPLET(card)) {
 		if (env->flags & SC_SEC_ENV_KEY_REF_ASYMMETRIC) {
-			sc_error(card->ctx, "asymmetric keyref not supported.\n");
+			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "asymmetric keyref not supported.\n");
 			return SC_ERROR_NOT_SUPPORTED;
 		}
 		if (se_num > 0) {
-			sc_error(card->ctx, "restore security environment not supported.\n");
+			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "restore security environment not supported.\n");
 			return SC_ERROR_NOT_SUPPORTED;
 		}
 	}
@@ -633,18 +628,20 @@ static int setcos_set_security_env2(sc_card_t *card,
 	apdu.resplen = 0;
 	if (se_num > 0) {
 		r = sc_lock(card);
-		SC_TEST_RET(card->ctx, r, "sc_lock() failed");
+		SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "sc_lock() failed");
 		locked = 1;
 	}
 	if (apdu.datalen != 0) {
 		r = sc_transmit_apdu(card, &apdu);
 		if (r) {
-			sc_perror(card->ctx, r, "APDU transmit failed");
+			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+				"%s: APDU transmit failed", sc_strerror(r));
 			goto err;
 		}
 		r = sc_check_sw(card, apdu.sw1, apdu.sw2);
 		if (r) {
-			sc_perror(card->ctx, r, "Card returned error");
+			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+				"%s: Card returned error", sc_strerror(r));
 			goto err;
 		}
 	}
@@ -653,7 +650,7 @@ static int setcos_set_security_env2(sc_card_t *card,
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0x22, 0xF2, se_num);
 	r = sc_transmit_apdu(card, &apdu);
 	sc_unlock(card);
-	SC_TEST_RET(card->ctx, r, "APDU transmit failed");
+	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
 	return sc_check_sw(card, apdu.sw1, apdu.sw2);
 err:
 	if (locked)
@@ -671,7 +668,7 @@ static int setcos_set_security_env(sc_card_t *card,
 		tmp.flags &= ~SC_SEC_ENV_ALG_PRESENT;
 		tmp.flags |= SC_SEC_ENV_ALG_REF_PRESENT;
 		if (tmp.algorithm != SC_ALGORITHM_RSA) {
-			sc_error(card->ctx, "Only RSA algorithm supported.\n");
+			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "Only RSA algorithm supported.\n");
 			return SC_ERROR_NOT_SUPPORTED;
 		}
 		switch (card->type) {
@@ -684,7 +681,7 @@ static int setcos_set_security_env(sc_card_t *card,
 		case SC_CARD_TYPE_SETCOS_EID_V2_1:
 			break;
 		default:
-			sc_error(card->ctx, "Card does not support RSA.\n");
+			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "Card does not support RSA.\n");
 			return SC_ERROR_NOT_SUPPORTED;
 			break;
 		}
@@ -932,7 +929,7 @@ static int setcos_list_files(sc_card_t *card, u8 * buf, size_t buflen)
 	apdu.resplen = buflen;
 	apdu.le = buflen > 256 ? 256 : buflen;
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, r, "APDU transmit failed");
+	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
 	if (card->type == SC_CARD_TYPE_SETCOS_44 && apdu.sw1 == 0x6A && apdu.sw2 == 0x82)
 		return 0; /* no files found */
 	if (apdu.resplen == 0)
@@ -965,7 +962,7 @@ static int setcos_putdata(struct sc_card *card, struct sc_cardctl_setcos_data_ob
 	int				r;
 	struct sc_apdu			apdu;
 
-	SC_FUNC_CALLED(card->ctx, 1);
+	SC_FUNC_CALLED(card->ctx, SC_LOG_DEBUG_VERBOSE);
 
 	memset(&apdu, 0, sizeof(apdu));
 	apdu.cse     = SC_APDU_CASE_3_SHORT;
@@ -978,12 +975,12 @@ static int setcos_putdata(struct sc_card *card, struct sc_cardctl_setcos_data_ob
 	apdu.data    = data_obj->Data;
 
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, r, "APDU transmit failed");
+	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
 
 	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
-	SC_TEST_RET(card->ctx, r, "PUT_DATA returned error");
+	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "PUT_DATA returned error");
 
-	SC_FUNC_RETURN(card->ctx, 1, r);
+	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, r);
 }
 
 /* Read internal data, e.g. get RSA public key */
@@ -992,7 +989,7 @@ static int setcos_getdata(struct sc_card *card, struct sc_cardctl_setcos_data_ob
 	int				r;
 	struct sc_apdu			apdu;
 
-	SC_FUNC_CALLED(card->ctx, 1);
+	SC_FUNC_CALLED(card->ctx, SC_LOG_DEBUG_VERBOSE);
 
 	memset(&apdu, 0, sizeof(apdu));
 	apdu.cse     = SC_APDU_CASE_2_SHORT;
@@ -1009,17 +1006,17 @@ static int setcos_getdata(struct sc_card *card, struct sc_cardctl_setcos_data_ob
 	apdu.resplen = data_obj->DataLen;
 
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, r, "APDU transmit failed");
+	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
 
 	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
-	SC_TEST_RET(card->ctx, r, "GET_DATA returned error");
+	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "GET_DATA returned error");
 
 	if (apdu.resplen > data_obj->DataLen)
 		r = SC_ERROR_WRONG_LENGTH;
 	else
 		data_obj->DataLen = apdu.resplen;
 
-	SC_FUNC_RETURN(card->ctx, 1, r);
+	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, r);
 }
 
 /* Generate or store a key */
@@ -1030,7 +1027,7 @@ static int setcos_generate_store_key(sc_card_t *card,
 	u8	sbuf[SC_MAX_APDU_BUFFER_SIZE];
 	int	r, len;
 
-	SC_FUNC_CALLED(card->ctx, 1);
+	SC_FUNC_CALLED(card->ctx, SC_LOG_DEBUG_VERBOSE);
 
 	/* Setup key-generation paramters */
 	len = 0;
@@ -1065,12 +1062,12 @@ static int setcos_generate_store_key(sc_card_t *card,
 	apdu.lc	= len;
 
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, r, "APDU transmit failed");
+	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
 
 	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
-	SC_TEST_RET(card->ctx, r, "STORE/GENERATE_KEY returned error");
+	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "STORE/GENERATE_KEY returned error");
 
-	SC_FUNC_RETURN(card->ctx, 1, r);
+	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, r);
 }
 
 static int setcos_activate_file(sc_card_t *card)
@@ -1083,12 +1080,12 @@ static int setcos_activate_file(sc_card_t *card)
 	apdu.data = sbuf;
 
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, r, "APDU transmit failed");
+	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
 
 	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
-	SC_TEST_RET(card->ctx, r, "ACTIVATE_FILE returned error");
+	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "ACTIVATE_FILE returned error");
 
-	SC_FUNC_RETURN(card->ctx, 1, r);
+	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, r);
 }
 
 static int setcos_card_ctl(sc_card_t *card, unsigned long cmd, void *ptr)
@@ -1122,7 +1119,6 @@ static struct sc_card_driver *sc_get_driver(void)
 	setcos_ops = *iso_drv->ops;
 	setcos_ops.match_card = setcos_match_card;
 	setcos_ops.init = setcos_init;
-	setcos_ops.finish = setcos_finish;
 	if (iso_ops == NULL)
 		iso_ops = iso_drv->ops;
 	setcos_ops.create_file = setcos_create_file;
