@@ -25,7 +25,7 @@
 extern "C" {
 #endif
 
-#include <opensc/opensc.h>
+#include "libopensc/opensc.h"
 
 #define SC_PKCS15_CACHE_DIR		".eid"
 
@@ -34,6 +34,12 @@ extern "C" {
 #define SC_PKCS15_MAX_LABEL_SIZE	255
 #define SC_PKCS15_MAX_ID_SIZE		255
 
+/* When changing this value, change also initialisation of the 
+ * static ASN1 variables, that use this macro,
+ * like for example, 'c_asn1_access_control_rules' 
+ * in src/libopensc/asn1.c */
+#define SC_PKCS15_MAX_ACCESS_RULES      8
+
 struct sc_pkcs15_id {
 	u8 value[SC_PKCS15_MAX_ID_SIZE];
 	size_t len;
@@ -63,14 +69,20 @@ typedef struct sc_pkcs15_id sc_pkcs15_id_t;
 #define SC_PKCS15_PIN_TYPE_HALFNIBBLE_BCD		3
 #define SC_PKCS15_PIN_TYPE_ISO9564_1			4
 
+#define SC_PKCS15_PIN_AUTH_TYPE_PIN			0
+#define SC_PKCS15_PIN_AUTH_TYPE_AUTH_KEY		1
+#define SC_PKCS15_PIN_AUTH_TYPE_SM_KEY			2
+
 struct sc_pkcs15_pin_info {
 	struct sc_pkcs15_id auth_id;
 	int reference;
 	unsigned int flags, type;
+	unsigned int auth_method;
 	size_t min_length, stored_length, max_length;
 	u8 pad_char;
 	struct sc_path path;
 	int tries_left;
+	int max_tries;
 
 	unsigned int magic;
 };
@@ -85,11 +97,6 @@ typedef struct sc_pkcs15_pin_info sc_pkcs15_pin_info_t;
 #define SC_PKCS15_ALGO_OP_HASH			0x40
 #define SC_PKCS15_ALGO_OP_GENERATE_KEY		0x80
 
-struct sc_pkcs15_algorithm_info {
-	int reference;
-	int algorithm, supported_operations;
-};
-
 /* A large integer, big endian notation */
 struct sc_pkcs15_bignum {
 	u8 *		data;
@@ -142,9 +149,23 @@ struct sc_pkcs15_prkey_dsa {
 	sc_pkcs15_bignum_t priv;
 };
 
+	/* The ecParameters are kept in DER format
+	 * as certificates, and pkcs11 process them as DER 
+	 * If needed, they can be parsed
+	 */ 
+struct sc_pkcs15_pubkey_ec {
+	sc_pkcs15_der_t		ecparameters;
+	sc_pkcs15_der_t		ecpointQ; /* note this is der */
+	size_t 				field_length; /* in bits */
+};
+
+struct sc_pkcs15_prkey_ec {
+	sc_pkcs15_der_t     ecparameters;
+	sc_pkcs15_bignum_t	privateD; /* note this is bignum */
+};
+
 struct sc_pkcs15_pubkey_gostr3410 {
-	sc_pkcs15_bignum_t x;
-	sc_pkcs15_bignum_t y;
+	sc_pkcs15_bignum_t xy;
 };
 
 struct sc_pkcs15_prkey_gostr3410 {
@@ -154,11 +175,13 @@ struct sc_pkcs15_prkey_gostr3410 {
 
 struct sc_pkcs15_pubkey {
 	int algorithm;
+	struct sc_algorithm_id * alg_id;
 
 	/* Decoded key */
 	union {
 		struct sc_pkcs15_pubkey_rsa rsa;
 		struct sc_pkcs15_pubkey_dsa dsa;
+		struct sc_pkcs15_pubkey_ec ec;
 		struct sc_pkcs15_pubkey_gostr3410 gostr3410;
 	} u;
 
@@ -169,9 +192,12 @@ typedef struct sc_pkcs15_pubkey sc_pkcs15_pubkey_t;
 
 struct sc_pkcs15_prkey {
 	unsigned int algorithm;
+/* TODO do we need:	struct sc_algorithm_id * alg_id; */
+
 	union {
 		struct sc_pkcs15_prkey_rsa rsa;
 		struct sc_pkcs15_prkey_dsa dsa;
+		struct sc_pkcs15_prkey_ec ec;
 		struct sc_pkcs15_prkey_gostr3410 gostr3410;
 	} u;
 };
@@ -201,7 +227,7 @@ struct sc_pkcs15_cert {
 	u8 *crl;
 	size_t crl_len;
 
-	struct sc_pkcs15_pubkey key;
+	struct sc_pkcs15_pubkey * key;
 	u8 *data;	/* DER encoded raw cert */
 	size_t data_len;
 };
@@ -264,13 +290,37 @@ struct sc_pkcs15_keyinfo_gostparams
 	unsigned int gostr3410, gostr3411, gost28147;
 };
 
+/* AccessMode bit definitions specified in PKCS#15 v1.1
+ * and extended by IAS/ECC v1.0.1 specification. */
+#define SC_PKCS15_ACCESS_RULE_MODE_READ         0x01
+#define SC_PKCS15_ACCESS_RULE_MODE_UPDATE       0x02
+#define SC_PKCS15_ACCESS_RULE_MODE_EXECUTE      0x04
+#define SC_PKCS15_ACCESS_RULE_MODE_DELETE       0x08
+#define SC_PKCS15_ACCESS_RULE_MODE_ATTRIBUTE    0x10
+#define SC_PKCS15_ACCESS_RULE_MODE_PSO_CDS      0x20
+#define SC_PKCS15_ACCESS_RULE_MODE_PSO_VERIFY   0x40
+#define SC_PKCS15_ACCESS_RULE_MODE_PSO_DECRYPT  0x80
+#define SC_PKCS15_ACCESS_RULE_MODE_PSO_ENCRYPT  0x100
+#define SC_PKCS15_ACCESS_RULE_MODE_INT_AUTH     0x200
+#define SC_PKCS15_ACCESS_RULE_MODE_EXT_AUTH     0x400
+
+struct sc_pkcs15_accessrule {
+	unsigned access_mode;
+	struct sc_pkcs15_id auth_id;
+};
+typedef struct sc_pkcs15_accessrule sc_pkcs15_accessrule_t;
+
+
 struct sc_pkcs15_prkey_info {
 	struct sc_pkcs15_id id;	/* correlates to public certificate id */
 	unsigned int usage, access_flags;
 	int native, key_reference;
-	size_t modulus_length;
-	u8     *subject;
-	size_t subject_len;
+	/* convert to union if other types are supported */
+	size_t modulus_length; /* RSA */
+	size_t field_length;   /* EC in bits */
+
+	struct sc_pkcs15_der subject;
+
 	void   *params;
 	size_t params_len;
 
@@ -282,9 +332,12 @@ struct sc_pkcs15_pubkey_info {
 	struct sc_pkcs15_id id;	/* correlates to private key id */
 	unsigned int usage, access_flags;
 	int native, key_reference;
-	size_t modulus_length;
-	u8     *subject;
-	size_t subject_len;
+	/* convert to union if other types are supported */
+	size_t modulus_length; /* RSA */
+	size_t field_length;   /* EC in bits */
+
+	struct sc_pkcs15_der subject;
+
 	void   *params;
 	size_t params_len;
 
@@ -298,11 +351,13 @@ typedef struct sc_pkcs15_pubkey_info sc_pkcs15_pubkey_info_t;
 #define SC_PKCS15_TYPE_PRKEY_RSA		0x101
 #define SC_PKCS15_TYPE_PRKEY_DSA		0x102
 #define SC_PKCS15_TYPE_PRKEY_GOSTR3410		0x103
+#define SC_PKCS15_TYPE_PRKEY_EC		0x104
 
 #define SC_PKCS15_TYPE_PUBKEY			0x200
 #define SC_PKCS15_TYPE_PUBKEY_RSA		0x201
 #define SC_PKCS15_TYPE_PUBKEY_DSA		0x202
 #define SC_PKCS15_TYPE_PUBKEY_GOSTR3410		0x203
+#define SC_PKCS15_TYPE_PUBKEY_EC		0x204
 
 #define SC_PKCS15_TYPE_CERT			0x400
 #define SC_PKCS15_TYPE_CERT_X509		0x401
@@ -326,15 +381,21 @@ struct sc_pkcs15_object {
 	unsigned int flags;
 	struct sc_pkcs15_id auth_id;
 
+	int usage_counter;
 	int user_consent;
 
+	struct sc_pkcs15_accessrule access_rules[SC_PKCS15_MAX_ACCESS_RULES];
+
 	/* Object type specific data */
 	void *data;
+	/* emulated object pointer */
+	void *emulated;
+
 
 	struct sc_pkcs15_df *df; /* can be NULL, if object is 'floating' */
 	struct sc_pkcs15_object *next, *prev; /* used only internally */
 	
-	struct sc_pkcs15_der der;
+	struct sc_pkcs15_der content;
 };
 typedef struct sc_pkcs15_object sc_pkcs15_object_t;
 
@@ -350,6 +411,8 @@ typedef struct sc_pkcs15_object sc_pkcs15_object_t;
 #define SC_PKCS15_AODF			8
 #define SC_PKCS15_DF_TYPE_COUNT		9
 
+struct sc_pkcs15_card;
+
 struct sc_pkcs15_df {
 	struct sc_file *file;
 
@@ -372,14 +435,14 @@ typedef struct sc_pkcs15_unusedspace sc_pkcs15_unusedspace_t;
 
 #define SC_PKCS15_CARD_MAGIC		0x10203040
 
-typedef struct {
+typedef struct sc_pkcs15_sec_env_info {
 	int			se;
 	struct sc_object_id	owner;
 	u8			aid[SC_MAX_AID_SIZE];
 	size_t			aid_len;
 } sc_pkcs15_sec_env_info_t;
 
-typedef struct {
+typedef struct sc_pkcs15_tokeninfo {
 	unsigned int version;
 	unsigned int flags;
 	char *label;
@@ -389,47 +452,52 @@ typedef struct {
 	char *preferred_language;
 	sc_pkcs15_sec_env_info_t **seInfo;
 	size_t num_seInfo;
+
+	struct sc_supported_algo_info supported_algos[SC_MAX_SUPPORTED_ALGORITHMS];
 } sc_pkcs15_tokeninfo_t;
 
+struct sc_pkcs15_operations   {
+	int (*parse_df)(struct sc_pkcs15_card *, struct sc_pkcs15_df *);
+	void (*clear)(struct sc_pkcs15_card *);
+};
+
 typedef struct sc_pkcs15_card {
 	sc_card_t *card;
-	char *label;
-	/* fields from TokenInfo: */
-	int version;
-	char *serial_number, *manufacturer_id;
-	char *last_update;
 	unsigned int flags;
-	struct sc_pkcs15_algorithm_info alg_info[1];
 
 	sc_file_t *file_app;
 	sc_file_t *file_tokeninfo, *file_odf, *file_unusedspace;
 
 	struct sc_pkcs15_df *df_list;
 	struct sc_pkcs15_object *obj_list;
-	int record_lengths[SC_PKCS15_DF_TYPE_COUNT];
+	sc_pkcs15_tokeninfo_t *tokeninfo;
 	sc_pkcs15_unusedspace_t *unusedspace_list;
 	int unusedspace_read;
 
 	struct sc_pkcs15_card_opts {
-		int use_cache;
+		int use_file_cache;
+		int use_pin_cache;
+		int pin_cache_counter;
 	} opts;
 
-	sc_pkcs15_sec_env_info_t **seInfo;
-	size_t num_seInfo;
 
 	unsigned int magic;
 
 	void *dll_handle;		/* shared lib for emulated cards */
-	char *preferred_language;
+
+	struct sc_pkcs15_operations ops;
+
 } sc_pkcs15_card_t;
 
-#define SC_PKCS15_CARD_FLAG_READONLY		0x01
-#define SC_PKCS15_CARD_FLAG_LOGIN_REQUIRED	0x02
-#define SC_PKCS15_CARD_FLAG_PRN_GENERATION	0x04
-#define SC_PKCS15_CARD_FLAG_EID_COMPLIANT	0x08
-#define SC_PKCS15_CARD_FLAG_SIGN_WITH_DECRYPT	0x10000000
-#define SC_PKCS15_CARD_FLAG_EMULATED		0x20000000
-#define SC_PKCS15_CARD_FLAG_FIX_INTEGERS	0x40000000
+/* flags suitable for sc_pkcs15_tokeninfo_t */
+#define SC_PKCS15_TOKEN_READONLY			0x01
+#define SC_PKCS15_TOKEN_LOGIN_REQUIRED			0x02 /* Don't use */
+#define SC_PKCS15_TOKEN_PRN_GENERATION			0x04
+#define SC_PKCS15_TOKEN_EID_COMPLIANT			0x08
+
+/* flags suitable for sc_pkcs15_card_t */
+#define SC_PKCS15_CARD_FLAG_EMULATED			0x02000000
+#define SC_PKCS15_CARD_FLAG_FIX_INTEGERS		0x04000000
 
 /* sc_pkcs15_bind:  Binds a card object to a PKCS #15 card object
  * and initializes a new PKCS #15 card object.  Will return
@@ -465,30 +533,45 @@ int sc_pkcs15_compute_signature(struct sc_pkcs15_card *p15card,
 				unsigned long alg_flags, const u8 *in,
 				size_t inlen, u8 *out, size_t outlen);
 
-int sc_pkcs15_read_pubkey(struct sc_pkcs15_card *card,
-			const struct sc_pkcs15_object *obj,
-			struct sc_pkcs15_pubkey **out);
-int sc_pkcs15_decode_pubkey_rsa(struct sc_context *ctx,
-	       		struct sc_pkcs15_pubkey_rsa *pubkey,
+int sc_pkcs15_read_pubkey(struct sc_pkcs15_card *,
+			const struct sc_pkcs15_object *,
+			struct sc_pkcs15_pubkey **);
+int sc_pkcs15_decode_pubkey_rsa(struct sc_context *,
+	       		struct sc_pkcs15_pubkey_rsa *,
 			const u8 *, size_t);
 int sc_pkcs15_encode_pubkey_rsa(struct sc_context *,
 			struct sc_pkcs15_pubkey_rsa *, u8 **, size_t *);
-int sc_pkcs15_decode_pubkey_dsa(struct sc_context *ctx,
-	       		struct sc_pkcs15_pubkey_dsa *pubkey,
+int sc_pkcs15_decode_pubkey_dsa(struct sc_context *,
+	       		struct sc_pkcs15_pubkey_dsa *,
 			const u8 *, size_t);
 int sc_pkcs15_encode_pubkey_dsa(struct sc_context *,
 			struct sc_pkcs15_pubkey_dsa *, u8 **, size_t *);
+int sc_pkcs15_decode_pubkey_gostr3410(sc_context_t *,
+		struct sc_pkcs15_pubkey_gostr3410 *, const u8 *, size_t);
+int sc_pkcs15_encode_pubkey_gostr3410(sc_context_t *,
+		struct sc_pkcs15_pubkey_gostr3410 *, u8 **, size_t *);
+int sc_pkcs15_decode_pubkey_ec(struct sc_context *,
+			struct sc_pkcs15_pubkey_ec *, const u8 *, size_t);
+int sc_pkcs15_encode_pubkey_ec(struct sc_context *,
+				struct sc_pkcs15_pubkey_ec *, u8 **, size_t *);
 int sc_pkcs15_decode_pubkey(struct sc_context *,
 	       		struct sc_pkcs15_pubkey *, const u8 *, size_t);
 int sc_pkcs15_encode_pubkey(struct sc_context *,
 			struct sc_pkcs15_pubkey *, u8 **, size_t *);
-void sc_pkcs15_erase_pubkey(struct sc_pkcs15_pubkey *pubkey);
-void sc_pkcs15_free_pubkey(struct sc_pkcs15_pubkey *pubkey);
-
-int sc_pkcs15_read_prkey(struct sc_pkcs15_card *card,
-			const struct sc_pkcs15_object *obj,
+void sc_pkcs15_erase_pubkey(struct sc_pkcs15_pubkey *);
+void sc_pkcs15_free_pubkey(struct sc_pkcs15_pubkey *);
+int sc_pkcs15_pubkey_from_prvkey(struct sc_context *, struct sc_pkcs15_prkey *, 
+			struct sc_pkcs15_pubkey **);
+int sc_pkcs15_pubkey_from_cert(struct sc_context *, struct sc_pkcs15_der *, 
+			struct sc_pkcs15_pubkey **);
+int sc_pkcs15_pubkey_from_spki_filename(struct sc_context *, 
+			char *, sc_pkcs15_pubkey_t ** );
+int sc_pkcs15_pubkey_from_spki(struct sc_context *,
+			sc_pkcs15_pubkey_t **, u8 *, size_t, int);
+int sc_pkcs15_read_prkey(struct sc_pkcs15_card *,
+			const struct sc_pkcs15_object *,
 			const char *passphrase,
-			struct sc_pkcs15_prkey **out);
+			struct sc_pkcs15_prkey **);
 int sc_pkcs15_decode_prkey(struct sc_context *,
 			struct sc_pkcs15_prkey *,
 			const u8 *, size_t);
@@ -543,14 +626,14 @@ int sc_pkcs15_find_pubkey_by_id(struct sc_pkcs15_card *card,
 			       struct sc_pkcs15_object **out);
 
 int sc_pkcs15_verify_pin(struct sc_pkcs15_card *card,
-			 struct sc_pkcs15_pin_info *pin,
+			 struct sc_pkcs15_object *pin_obj,
 			 const u8 *pincode, size_t pinlen);
 int sc_pkcs15_change_pin(struct sc_pkcs15_card *card,
-			 struct sc_pkcs15_pin_info *pin,
+			 struct sc_pkcs15_object *pin_obj,
 			 const u8 *oldpincode, size_t oldpinlen,
 			 const u8 *newpincode, size_t newpinlen);
 int sc_pkcs15_unblock_pin(struct sc_pkcs15_card *card,
-			 struct sc_pkcs15_pin_info *pin,
+			 struct sc_pkcs15_object *pin_obj,
 			 const u8 *puk, size_t puklen,
 			 const u8 *newpin, size_t newpinlen);
 int sc_pkcs15_find_pin_by_auth_id(struct sc_pkcs15_card *card,
@@ -559,9 +642,19 @@ int sc_pkcs15_find_pin_by_auth_id(struct sc_pkcs15_card *card,
 int sc_pkcs15_find_pin_by_reference(struct sc_pkcs15_card *card,
 				    const sc_path_t *path, int reference,
 				    struct sc_pkcs15_object **out);
+int sc_pkcs15_find_pin_by_type_and_reference(struct sc_pkcs15_card *card,
+				    const sc_path_t *path, unsigned auth_method, 
+				    int reference,
+				    struct sc_pkcs15_object **out);
 int sc_pkcs15_find_so_pin(struct sc_pkcs15_card *card,
 			struct sc_pkcs15_object **out);
 
+void sc_pkcs15_pincache_add(struct sc_pkcs15_card *, struct sc_pkcs15_object *, 
+			const u8 *, size_t);
+int sc_pkcs15_pincache_revalidate(struct sc_pkcs15_card *p15card, 
+			const sc_pkcs15_object_t *obj);
+void sc_pkcs15_pincache_clear(struct sc_pkcs15_card *p15card);
+
 int sc_pkcs15_encode_dir(struct sc_context *ctx,
 			struct sc_pkcs15_card *card,
 			u8 **buf, size_t *buflen);
@@ -625,9 +718,8 @@ int sc_pkcs15_add_object(struct sc_pkcs15_card *p15card,
 			 struct sc_pkcs15_object *obj);
 void sc_pkcs15_remove_object(struct sc_pkcs15_card *p15card,
 			     struct sc_pkcs15_object *obj);
-int sc_pkcs15_add_df(struct sc_pkcs15_card *p15card,
-		     unsigned int type, const sc_path_t *path,
-		     const struct sc_file *file);
+int sc_pkcs15_add_df(struct sc_pkcs15_card *, unsigned int, 
+			const sc_path_t *, const struct sc_file *);
 void sc_pkcs15_remove_df(struct sc_pkcs15_card *p15card,
 			 struct sc_pkcs15_df *df);
 
@@ -679,10 +771,15 @@ const char *sc_pkcs15_print_id(const struct sc_pkcs15_id *id);
 void sc_pkcs15_format_id(const char *id_in, struct sc_pkcs15_id *id_out);
 int sc_pkcs15_hex_string_to_id(const char *in, struct sc_pkcs15_id *out);
 void sc_der_copy(sc_pkcs15_der_t *, const sc_pkcs15_der_t *);
-void sc_der_clear(sc_pkcs15_der_t *);
 /* Prepend 'parent' to 'child' in case 'child' is a relative path */
 int sc_pkcs15_make_absolute_path(const sc_path_t *parent, sc_path_t *child);
 
+/* Clean and free object content */
+void sc_pkcs15_free_object_content(struct sc_pkcs15_object *);
+/* Allocate and set object content */
+int sc_pkcs15_allocate_object_content(struct sc_pkcs15_object *,
+		const unsigned char *, size_t);
+
 /* New object search API.
  * More complex, but also more powerful.
  */
@@ -715,8 +812,8 @@ typedef struct sc_pkcs15emu_opt {
 extern int sc_pkcs15_bind_synthetic(sc_pkcs15_card_t *);
 extern int sc_pkcs15_is_emulation_only(sc_card_t *);
 
-int sc_pkcs15emu_object_add(sc_pkcs15_card_t *p15card, unsigned int type,
-			const sc_pkcs15_object_t *obj, const void *data);
+int sc_pkcs15emu_object_add(sc_pkcs15_card_t *, unsigned int,
+			const sc_pkcs15_object_t *, const void *);
 /* some wrapper functions for sc_pkcs15emu_object_add */
 int sc_pkcs15emu_add_pin_obj(sc_pkcs15_card_t *,
 	const sc_pkcs15_object_t *, const sc_pkcs15_pin_info_t *);
@@ -724,9 +821,13 @@ int sc_pkcs15emu_add_rsa_prkey(sc_pkcs15_card_t *,
 	const sc_pkcs15_object_t *, const sc_pkcs15_prkey_info_t *);
 int sc_pkcs15emu_add_rsa_pubkey(sc_pkcs15_card_t *,
 	const sc_pkcs15_object_t *, const sc_pkcs15_pubkey_info_t *);
-int sc_pkcs15emu_add_x509_cert(sc_pkcs15_card_t *p15card,
+int sc_pkcs15emu_add_ec_prkey(sc_pkcs15_card_t *,
+	const sc_pkcs15_object_t *, const sc_pkcs15_prkey_info_t *);
+int sc_pkcs15emu_add_ec_pubkey(sc_pkcs15_card_t *,
+	const sc_pkcs15_object_t *, const sc_pkcs15_pubkey_info_t *);
+int sc_pkcs15emu_add_x509_cert(sc_pkcs15_card_t *,
 	const sc_pkcs15_object_t *, const sc_pkcs15_cert_info_t *);
-int sc_pkcs15emu_add_data_object(sc_pkcs15_card_t *p15card,
+int sc_pkcs15emu_add_data_object(sc_pkcs15_card_t *,
 	const sc_pkcs15_object_t *, const sc_pkcs15_data_info_t *);
 
 #ifdef __cplusplus
