@@ -18,15 +18,18 @@
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
-#include "internal.h"
-#include "ctbcs.h"
+#include "config.h"
+
+#ifdef ENABLE_CTAPI
 #include <assert.h>
 #include <stdlib.h>
 #include <string.h>
 #include <ltdl.h>
 
+#include "internal.h"
+#include "ctbcs.h"
+
 #define GET_PRIV_DATA(r) ((struct ctapi_private_data *) (r)->drv_data)
-#define GET_SLOT_DATA(r) ((struct ctapi_slot_data *) (r)->drv_data)
 
 #ifdef _WIN32
 typedef char pascal CT_INIT_TYPE(unsigned short ctn, unsigned short Pn);
@@ -71,14 +74,11 @@ struct ctapi_private_data {
 	struct ctapi_functions funcs;
 	unsigned short ctn;
 	int ctapi_functional_units;
+	int slot;
 };
 
-struct ctapi_slot_data {
-	void *filler;
-};
-
-/* Reset slot or reader */
-static int ctapi_reset(sc_reader_t *reader, sc_slot_info_t *slot)
+/* Reset reader */
+static int ctapi_reset(sc_reader_t *reader)
 {
 	struct ctapi_private_data *priv = GET_PRIV_DATA(reader);
 	char rv;
@@ -87,7 +87,7 @@ static int ctapi_reset(sc_reader_t *reader, sc_slot_info_t *slot)
 
 	cmd[0] = CTBCS_CLA;
 	cmd[1] = CTBCS_INS_RESET;
-	cmd[2] = slot ? CTBCS_P1_INTERFACE1 + slot->id : CTBCS_P1_CT_KERNEL;
+	cmd[2] = priv->slot ? CTBCS_P1_INTERFACE1 + priv->slot : CTBCS_P1_CT_KERNEL;
 	cmd[3] = 0x00; /* No response. We might also use 0x01 (return ATR) or 0x02 (return historical bytes) here */
 	cmd[4] = 0x00;
 	dad = 1;
@@ -96,148 +96,18 @@ static int ctapi_reset(sc_reader_t *reader, sc_slot_info_t *slot)
 
 	rv = priv->funcs.CT_data(priv->ctn, &dad, &sad, 5, cmd, &lr, rbuf);
 	if (rv || (lr < 2)) {
-		sc_error(reader->ctx, "Error getting status of terminal: %d, using defaults\n", rv);
+		sc_debug(reader->ctx, SC_LOG_DEBUG_NORMAL, "Error getting status of terminal: %d, using defaults\n", rv);
 		return SC_ERROR_TRANSMIT_FAILED;
 	}
 	if (rbuf[lr-2] != 0x90) {
-		sc_error(reader->ctx, "SW1/SW2: 0x%x/0x%x\n", rbuf[lr-2], rbuf[lr-1]);
+		sc_debug(reader->ctx, SC_LOG_DEBUG_NORMAL, "SW1/SW2: 0x%x/0x%x\n", rbuf[lr-2], rbuf[lr-1]);
 		return SC_ERROR_TRANSMIT_FAILED;
 	}
 	return 0;
 }
 
-static void set_default_fu(sc_reader_t *reader)
-{
-	if (!reader) return;
 
-	reader->slot_count = 1;
-	reader->slot[0].id = 0;
-	reader->slot[0].capabilities = 0;
-	reader->slot[0].atr_len = 0;
-	reader->slot[0].drv_data = NULL;
-}
-
-/* Detect functional units of the reader according to CT-BCS spec version 1.0
-   (14.04.2004, http://www.teletrust.de/down/mct1-0_t4.zip) */
-static void detect_functional_units(sc_reader_t *reader)
-{
-	struct ctapi_private_data *priv = GET_PRIV_DATA(reader);
-	char rv;
-	u8 cmd[5], rbuf[256], sad, dad;
-	unsigned short lr;
-	int NumUnits;
-	int i;
-
-	priv->ctapi_functional_units = 0;
-
-	cmd[0] = CTBCS_CLA;
-	cmd[1] = CTBCS_INS_STATUS;
-	cmd[2] = CTBCS_P1_CT_KERNEL;
-	cmd[3] = CTBCS_P2_STATUS_TFU;
-	cmd[4] = 0x00;
-	dad = 1;
-	sad = 2;
-	lr = 256;
-
-	rv = priv->funcs.CT_data(priv->ctn, &dad, &sad, 5, cmd, &lr, rbuf);
-	if (rv || (lr < 4) || (rbuf[lr-2] != 0x90)) {
-		sc_error(reader->ctx, "Error getting status of terminal: %d, using defaults\n", rv);
-		set_default_fu(reader);
-		return;
-	}
-	if (rbuf[0] != CTBCS_P2_STATUS_TFU) {
-		/* Number of slots might also detected by using CTBCS_P2_STATUS_ICC.
-		   If you think that's important please do it... ;) */
-		set_default_fu(reader);
-		sc_error(reader->ctx, "Invalid data object returnd on CTBCS_P2_STATUS_TFU: 0x%x\n", rbuf[0]);
-		return;
-	}
-	NumUnits = rbuf[1];
-	if (NumUnits + 4 > lr) {
-		set_default_fu(reader);
-		sc_error(reader->ctx, "Invalid data returnd: %d functional units, size %d\n", NumUnits, rv);
-		set_default_fu(reader);
-		return;
-	}
-	reader->slot_count = 0;
-	for(i = 0; i < NumUnits; i++) {
-		switch(rbuf[i+2])
-		{
-			case CTBCS_P1_INTERFACE1:
-			case CTBCS_P1_INTERFACE2:
-			case CTBCS_P1_INTERFACE3:
-			case CTBCS_P1_INTERFACE4:
-			case CTBCS_P1_INTERFACE5:
-			case CTBCS_P1_INTERFACE6:
-			case CTBCS_P1_INTERFACE7:
-			case CTBCS_P1_INTERFACE8:
-			case CTBCS_P1_INTERFACE9:
-			case CTBCS_P1_INTERFACE10:
-			case CTBCS_P1_INTERFACE11:
-			case CTBCS_P1_INTERFACE12:
-			case CTBCS_P1_INTERFACE13:
-			case CTBCS_P1_INTERFACE14:
-			/* Maybe a weak point here if multiple interfaces are present and not returned
-			   in the "canonical" order. This is not forbidden by the specs, but why should
-			   anyone want to do that? */
-				if (reader->slot_count >= SC_MAX_SLOTS) {
-					sc_debug(reader->ctx, "Ignoring slot id 0x%x, can only handle %d slots\n", rbuf[i+2], SC_MAX_SLOTS);
-				} else {
-					reader->slot[reader->slot_count].id = reader->slot_count;
-					reader->slot[reader->slot_count].capabilities = 0; /* Just to start with */
-					reader->slot[reader->slot_count].atr_len = 0;
-					reader->slot[reader->slot_count].drv_data = NULL;
-					reader->slot_count++;
-				}
-				break;
-
-			case CTBCS_P1_DISPLAY:
-				priv->ctapi_functional_units |= CTAPI_FU_DISPLAY;
-				sc_debug(reader->ctx, "Display detected\n");
-				break;
-
-			case CTBCS_P1_KEYPAD:
-				priv->ctapi_functional_units |= CTAPI_FU_KEYBOARD;
-				sc_debug(reader->ctx, "Keypad detected\n");
-				break;
-
-			case CTBCS_P1_PRINTER:
-				priv->ctapi_functional_units |= CTAPI_FU_PRINTER;
-				sc_debug(reader->ctx, "Printer detected\n");
-				break;
-
-			case CTBCS_P1_FINGERPRINT:
-			case CTBCS_P1_VOICEPRINT:
-			case CTBCS_P1_DSV:
-			case CTBCS_P1_FACE_RECOGNITION:
-			case CTBCS_P1_IRISSCAN:
-				priv->ctapi_functional_units |= CTAPI_FU_BIOMETRIC;
-				sc_debug(reader->ctx, "Biometric sensor detected\n");
-				break;
-
-			default:
-				sc_debug(reader->ctx, "Unknown functional unit 0x%x\n", rbuf[i+2]);
-		}
-
-	}
-	if (reader->slot_count == 0) {
-		sc_debug(reader->ctx, "No slots returned, assuming one default slot\n");
-		set_default_fu(reader);
-	}
-	/* CT-BCS does not define Keyboard/Display for each slot, so I assume
-	   those additional units can be used for each slot */
-	if (priv->ctapi_functional_units) {
-		for(i = 0; i < reader->slot_count; i++)	{
-			if (priv->ctapi_functional_units & CTAPI_FU_KEYBOARD)
-			reader->slot[i].capabilities |= SC_SLOT_CAP_PIN_PAD;
-			if (priv->ctapi_functional_units & CTAPI_FU_DISPLAY)
-			reader->slot[i].capabilities |= SC_SLOT_CAP_DISPLAY;
-		}
-	}
-}
-
-static int refresh_slot_attributes(sc_reader_t *reader,
-				   sc_slot_info_t *slot)
+static int refresh_attributes(sc_reader_t *reader)
 {
 	struct ctapi_private_data *priv = GET_PRIV_DATA(reader);
 	char rv;
@@ -253,39 +123,31 @@ static int refresh_slot_attributes(sc_reader_t *reader,
 	sad = 2;
 	lr = 256;
 
-	slot->flags = 0;
+	reader->flags = 0;
 
 	rv = priv->funcs.CT_data(priv->ctn, &dad, &sad, 5, cmd, &lr, rbuf);
 	if (rv || (lr < 3) || (rbuf[lr-2] != 0x90)) {
-		sc_error(reader->ctx, "Error getting status of terminal: %d/%d/0x%x\n", rv, lr, rbuf[lr-2]);
+		sc_debug(reader->ctx, SC_LOG_DEBUG_NORMAL, "Error getting status of terminal: %d/%d/0x%x\n", rv, lr, rbuf[lr-2]);
 		return SC_ERROR_TRANSMIT_FAILED;
 	}
 	if (lr < 4) {
-		/* Looks like older readers do not return data tag and length field, so assume one slot only */
-		if (slot->id > 0) {
-			sc_error(reader->ctx, "Status for slot id %d not returned, have only 1\n", slot->id);
-			return SC_ERROR_SLOT_NOT_FOUND;
-		}
 		if (rbuf[0] & CTBCS_DATA_STATUS_CARD)
-			slot->flags = SC_SLOT_CARD_PRESENT;
+			reader->flags = SC_READER_CARD_PRESENT;
 	} else {
 		if (rbuf[0] != CTBCS_P2_STATUS_ICC) {
 			/* Should we be more tolerant here? I do not think so... */
-			sc_error(reader->ctx, "Invalid data object returnd on CTBCS_P2_STATUS_ICC: 0x%x\n", rbuf[0]);
+			sc_debug(reader->ctx, SC_LOG_DEBUG_NORMAL, "Invalid data object returnd on CTBCS_P2_STATUS_ICC: 0x%x\n", rbuf[0]);
 		return SC_ERROR_TRANSMIT_FAILED;
-	}
-		if (rbuf[1] <= slot->id) {
-			sc_error(reader->ctx, "Status for slot id %d not returned, only %d\n", slot->id, rbuf[1]);
-			return SC_ERROR_SLOT_NOT_FOUND;
 		}
-		if (rbuf[2+slot->id] & CTBCS_DATA_STATUS_CARD)
-		slot->flags = SC_SLOT_CARD_PRESENT;
+		/* Fixme - should not be reached */
+		sc_debug(reader->ctx, SC_LOG_DEBUG_NORMAL, "Returned status for  %d slots\n", rbuf[1]);
+		reader->flags = SC_READER_CARD_PRESENT;
 	}
 
 	return 0;
 }
 
-static int ctapi_internal_transmit(sc_reader_t *reader, sc_slot_info_t *slot,
+static int ctapi_internal_transmit(sc_reader_t *reader,
 			 const u8 *sendbuf, size_t sendsize,
 			 u8 *recvbuf, size_t *recvsize,
 			 unsigned long control)
@@ -297,16 +159,15 @@ static int ctapi_internal_transmit(sc_reader_t *reader, sc_slot_info_t *slot,
 	
 	if (control)
 		dad = 1;
-	else if (!slot || slot->id == 0)
-		dad = 0;
 	else
-		dad = slot->id + 1; /* Adressing of multiple slots, according to CT API 1.0 */
+		dad = 0;
+
 	sad = 2;
 	lr = *recvsize;
 	
 	rv = priv->funcs.CT_data(priv->ctn, &dad, &sad, (unsigned short)sendsize, (u8 *) sendbuf, &lr, recvbuf);
 	if (rv != 0) {
-		sc_error(reader->ctx, "Error transmitting APDU: %d\n", rv);
+		sc_debug(reader->ctx, SC_LOG_DEBUG_NORMAL, "Error transmitting APDU: %d\n", rv);
 		return SC_ERROR_TRANSMIT_FAILED;
 	}
 	*recvsize = lr;
@@ -314,8 +175,7 @@ static int ctapi_internal_transmit(sc_reader_t *reader, sc_slot_info_t *slot,
 	return 0;
 }
 
-static int ctapi_transmit(sc_reader_t *reader, sc_slot_info_t *slot,
-	sc_apdu_t *apdu)
+static int ctapi_transmit(sc_reader_t *reader, sc_apdu_t *apdu)
 {
 	size_t       ssize, rsize, rbuflen = 0;
 	u8           *sbuf = NULL, *rbuf = NULL;
@@ -324,24 +184,22 @@ static int ctapi_transmit(sc_reader_t *reader, sc_slot_info_t *slot,
 	rsize = rbuflen = apdu->resplen + 2;
 	rbuf     = malloc(rbuflen);
 	if (rbuf == NULL) {
-		r = SC_ERROR_MEMORY_FAILURE;
+		r = SC_ERROR_OUT_OF_MEMORY;
 		goto out;
 	}
 	/* encode and log the APDU */
 	r = sc_apdu_get_octets(reader->ctx, apdu, &sbuf, &ssize, SC_PROTO_RAW);
 	if (r != SC_SUCCESS)
 		goto out;
-	if (reader->ctx->debug >= 6)
-		sc_apdu_log(reader->ctx, sbuf, ssize, 1);
-	r = ctapi_internal_transmit(reader, slot, sbuf, ssize,
+	sc_apdu_log(reader->ctx, SC_LOG_DEBUG_NORMAL, sbuf, ssize, 1);
+	r = ctapi_internal_transmit(reader, sbuf, ssize,
 					rbuf, &rsize, apdu->control);
 	if (r < 0) {
 		/* unable to transmit ... most likely a reader problem */
-		sc_error(reader->ctx, "unable to transmit");
+		sc_debug(reader->ctx, SC_LOG_DEBUG_NORMAL, "unable to transmit");
 		goto out;
 	}
-	if (reader->ctx->debug >= 6)
-		sc_apdu_log(reader->ctx, rbuf, rsize, 0);
+	sc_apdu_log(reader->ctx, SC_LOG_DEBUG_NORMAL, rbuf, rsize, 0);
 	/* set response */
 	r = sc_apdu_set_resp(reader->ctx, apdu, rbuf, rsize);
 out:
@@ -357,17 +215,17 @@ out:
 	return r;
 }
 
-static int ctapi_detect_card_presence(sc_reader_t *reader, sc_slot_info_t *slot)
+static int ctapi_detect_card_presence(sc_reader_t *reader)
 {
 	int r;
 	
-	r = refresh_slot_attributes(reader, slot);
+	r = refresh_attributes(reader);
 	if (r)
 		return r;
-	return slot->flags;
+	return reader->flags;
 }
 
-static int ctapi_connect(sc_reader_t *reader, sc_slot_info_t *slot)
+static int ctapi_connect(sc_reader_t *reader)
 {
 	struct ctapi_private_data *priv = GET_PRIV_DATA(reader);
 	char rv;
@@ -377,7 +235,7 @@ static int ctapi_connect(sc_reader_t *reader, sc_slot_info_t *slot)
 
 	cmd[0] = CTBCS_CLA;
 	cmd[1] = CTBCS_INS_REQUEST;
-	cmd[2] = CTBCS_P1_INTERFACE1+slot->id;
+	cmd[2] = CTBCS_P1_INTERFACE1;
 	cmd[3] = CTBCS_P2_REQUEST_GET_ATR;
 	cmd[4] = 0x00;
 	dad = 1;
@@ -386,26 +244,26 @@ static int ctapi_connect(sc_reader_t *reader, sc_slot_info_t *slot)
 
 	rv = priv->funcs.CT_data(priv->ctn, &dad, &sad, 5, cmd, &lr, rbuf);
 	if (rv || rbuf[lr-2] != 0x90) {
-		sc_error(reader->ctx, "Error activating card: %d\n", rv);
+		sc_debug(reader->ctx, SC_LOG_DEBUG_NORMAL, "Error activating card: %d\n", rv);
 		return SC_ERROR_TRANSMIT_FAILED;
 	}
 	if (lr < 2)
-		SC_FUNC_RETURN(reader->ctx, 0, SC_ERROR_INTERNAL);
+		SC_FUNC_RETURN(reader->ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_INTERNAL);
 	lr -= 2;
 	if (lr > SC_MAX_ATR_SIZE)
-		lr = SC_MAX_ATR_SIZE;
-	memcpy(slot->atr, rbuf, lr);
-	slot->atr_len = lr;
-	r = _sc_parse_atr(reader->ctx, slot);
+		return SC_ERROR_INTERNAL;
+	memcpy(reader->atr, rbuf, lr);
+	reader->atr_len = lr;
+	r = _sc_parse_atr(reader);
 
 #if 0	
-	if (slot->atr_info.Fi > 0) {
+	if (reader->atr_info.Fi > 0) {
 		/* Perform PPS negotiation */
 		cmd[1] = CTBCS_INS_RESET;
 		cmd[4] = 0x03;
 		cmd[5] = 0xFF;
 		cmd[6] = 0x10;
-		cmd[7] = (slot->atr_info.FI << 4) | slot->atr_info.DI;
+		cmd[7] = (reader->atr_info.FI << 4) | reader->atr_info.DI;
 		cmd[8] = 0x00;
 		dad = 1;
 		sad = 2;
@@ -413,7 +271,7 @@ static int ctapi_connect(sc_reader_t *reader, sc_slot_info_t *slot)
 
 		rv = priv->funcs.CT_data(priv->ctn, &dad, &sad, 9, cmd, &lr, rbuf);
 		if (rv) {
-			sc_error(reader->ctx, "Error negotiating PPS: %d\n", rv);
+			sc_debug(reader->ctx, SC_LOG_DEBUG_NORMAL, "Error negotiating PPS: %d\n", rv);
 			return SC_ERROR_TRANSMIT_FAILED;
 		}
 	}
@@ -421,17 +279,17 @@ static int ctapi_connect(sc_reader_t *reader, sc_slot_info_t *slot)
 	return 0;
 }
 
-static int ctapi_disconnect(sc_reader_t *reader, sc_slot_info_t *slot)
+static int ctapi_disconnect(sc_reader_t *reader)
 {
 	return 0;
 }
 
-static int ctapi_lock(sc_reader_t *reader, sc_slot_info_t *slot)
+static int ctapi_lock(sc_reader_t *reader)
 {
 	return 0;
 }
 
-static int ctapi_unlock(sc_reader_t *reader, sc_slot_info_t *slot)
+static int ctapi_unlock(sc_reader_t *reader)
 {
 	return 0;
 }
@@ -479,18 +337,22 @@ static int ctapi_load_module(sc_context_t *ctx,
 	struct ctapi_module *mod;
 	const scconf_list *list;
 	void *dlh;
-	int r, i;
+	int r, i, NumUnits;
+	u8 cmd[5], rbuf[256], sad, dad;
+	unsigned short lr;
+
+	
 	
 	list = scconf_find_list(conf, "ports");
 	if (list == NULL) {
-		sc_error(ctx, "No ports configured.\n");
+		sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "No ports configured.\n");
 		return -1;
 	}
 
 	val = conf->name->data;
 	dlh = lt_dlopen(val);
 	if (!dlh) {
-		sc_error(ctx, "Unable to open shared library '%s': %s\n", val, lt_dlerror());
+		sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "Unable to open shared library '%s': %s\n", val, lt_dlerror());
 		return -1;
 	}
 
@@ -513,16 +375,19 @@ static int ctapi_load_module(sc_context_t *ctx,
 		struct ctapi_private_data *priv;
 		
 		if (sscanf(list->data, "%d", &port) != 1) {
-			sc_error(ctx, "Port '%s' is not a number.\n", list->data);
+			sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "Port '%s' is not a number.\n", list->data);
 			continue;
 		}
 		rv = funcs.CT_init((unsigned short)mod->ctn_count, (unsigned short)port);
 		if (rv) {
-			sc_error(ctx, "CT_init() failed with %d\n", rv);
+			sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "CT_init() failed with %d\n", rv);
 			continue;
 		}
-		reader = (sc_reader_t *) calloc(1, sizeof(sc_reader_t));
-		priv = (struct ctapi_private_data *) malloc(sizeof(struct ctapi_private_data));
+		
+		reader = calloc(1, sizeof(sc_reader_t));
+		priv = calloc(1, sizeof(struct ctapi_private_data));
+		if (!priv)
+			return SC_ERROR_OUT_OF_MEMORY;
 		reader->drv_data = priv;
 		reader->ops = &ctapi_ops;
 		reader->driver = &ctapi_drv;
@@ -538,33 +403,112 @@ static int ctapi_load_module(sc_context_t *ctx,
 			free(reader);
 			break;
 		}
-		/* slot count and properties are set in detect_functional_units */
-		detect_functional_units(reader);
 		
-		ctapi_reset(reader, NULL);
-		for(i = 0; i < reader->slot_count; i++) {
-			refresh_slot_attributes(reader, &(reader->slot[i]));
+		/* Detect functional units of the reader according to CT-BCS spec version 1.0 
+		(14.04.2004, http://www.teletrust.de/down/mct1-0_t4.zip) */	
+		cmd[0] = CTBCS_CLA;
+		cmd[1] = CTBCS_INS_STATUS;
+		cmd[2] = CTBCS_P1_CT_KERNEL;
+		cmd[3] = CTBCS_P2_STATUS_TFU;
+		cmd[4] = 0x00;
+		dad = 1;
+		sad = 2;
+		lr = 256;
+		
+		rv = priv->funcs.CT_data(priv->ctn, &dad, &sad, 5, cmd, &lr, rbuf);
+		if (rv || (lr < 4) || (rbuf[lr-2] != 0x90)) {
+			sc_debug(reader->ctx, SC_LOG_DEBUG_NORMAL, "Error getting status of terminal: %d, using defaults\n", rv);
+		}
+		if (rbuf[0] != CTBCS_P2_STATUS_TFU) {
+			/* Number of slots might also detected by using CTBCS_P2_STATUS_ICC.
+			   If you think that's important please do it... ;) */
+			sc_debug(reader->ctx, SC_LOG_DEBUG_NORMAL, "Invalid data object returnd on CTBCS_P2_STATUS_TFU: 0x%x\n", rbuf[0]);
+		}
+		NumUnits = rbuf[1];
+		if (NumUnits + 4 > lr) {
+			sc_debug(reader->ctx, SC_LOG_DEBUG_NORMAL, "Invalid data returnd: %d functional units, size %d\n", NumUnits, rv);
+		}
+		priv->ctapi_functional_units = 0;
+		for(i = 0; i < NumUnits; i++) {
+			switch(rbuf[i+2]) {
+				case CTBCS_P1_INTERFACE1:
+				case CTBCS_P1_INTERFACE2:
+				case CTBCS_P1_INTERFACE3:
+				case CTBCS_P1_INTERFACE4:
+				case CTBCS_P1_INTERFACE5:
+				case CTBCS_P1_INTERFACE6:
+				case CTBCS_P1_INTERFACE7:
+				case CTBCS_P1_INTERFACE8:
+				case CTBCS_P1_INTERFACE9:
+				case CTBCS_P1_INTERFACE10:
+				case CTBCS_P1_INTERFACE11:
+				case CTBCS_P1_INTERFACE12:
+				case CTBCS_P1_INTERFACE13:
+				case CTBCS_P1_INTERFACE14:
+				/* Maybe a weak point here if multiple interfaces are present and not returned
+				   in the "canonical" order. This is not forbidden by the specs, but why should
+				   anyone want to do that? */
+					sc_debug(reader->ctx, SC_LOG_DEBUG_NORMAL, "Found slot id 0x%x\n", rbuf[i+2]);
+					break;
+
+				case CTBCS_P1_DISPLAY:
+					priv->ctapi_functional_units |= CTAPI_FU_DISPLAY;
+					sc_debug(reader->ctx, SC_LOG_DEBUG_NORMAL, "Display detected\n");
+					break;
+
+				case CTBCS_P1_KEYPAD:
+					priv->ctapi_functional_units |= CTAPI_FU_KEYBOARD;
+					sc_debug(reader->ctx, SC_LOG_DEBUG_NORMAL, "Keypad detected\n");
+					break;
+
+				case CTBCS_P1_PRINTER:
+					priv->ctapi_functional_units |= CTAPI_FU_PRINTER;
+					sc_debug(reader->ctx, SC_LOG_DEBUG_NORMAL, "Printer detected\n");
+					break;
+
+				case CTBCS_P1_FINGERPRINT:
+				case CTBCS_P1_VOICEPRINT:
+				case CTBCS_P1_DSV:
+				case CTBCS_P1_FACE_RECOGNITION:
+				case CTBCS_P1_IRISSCAN:
+					priv->ctapi_functional_units |= CTAPI_FU_BIOMETRIC;
+					sc_debug(reader->ctx, SC_LOG_DEBUG_NORMAL, "Biometric sensor detected\n");
+					break;
+
+				default:
+					sc_debug(reader->ctx, SC_LOG_DEBUG_NORMAL, "Unknown functional unit 0x%x\n", rbuf[i+2]);
+			}
+		}
+		/* CT-BCS does not define Keyboard/Display for each slot, so I assume
+		those additional units can be used for each slot */
+		if (priv->ctapi_functional_units) {
+			if (priv->ctapi_functional_units & CTAPI_FU_KEYBOARD)
+				reader->capabilities |= SC_READER_CAP_PIN_PAD;
+			if (priv->ctapi_functional_units & CTAPI_FU_DISPLAY)
+				reader->capabilities |= SC_READER_CAP_DISPLAY;
 		}
 		
+		ctapi_reset(reader);
+		refresh_attributes(reader);
 		mod->ctn_count++;
 	}
 	return 0;
 symerr:
-	sc_error(ctx, "Unable to resolve CT-API symbols.\n");
+	sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "Unable to resolve CT-API symbols.\n");
 	lt_dlclose(dlh);
 	return -1;
 }
 
-static int ctapi_init(sc_context_t *ctx, void **reader_data)
+static int ctapi_init(sc_context_t *ctx)
 {
 	int i;
 	struct ctapi_global_private_data *gpriv;
 	scconf_block **blocks = NULL, *conf_block = NULL;
 
-	gpriv = (struct ctapi_global_private_data *) calloc(1, sizeof(struct ctapi_global_private_data));
+	gpriv = calloc(1, sizeof(struct ctapi_global_private_data));
 	if (gpriv == NULL)
 		return SC_ERROR_OUT_OF_MEMORY;
-	*reader_data = gpriv;
+	ctx->reader_drv_data = gpriv;
 	
 	for (i = 0; ctx->conf_blocks[i] != NULL; i++) {
 		blocks = scconf_find_blocks(ctx->conf, ctx->conf_blocks[i],
@@ -585,9 +529,9 @@ static int ctapi_init(sc_context_t *ctx, void **reader_data)
 	return 0;
 }
 
-static int ctapi_finish(sc_context_t *ctx, void *prv_data)
+static int ctapi_finish(sc_context_t *ctx)
 {
-	struct ctapi_global_private_data *priv = (struct ctapi_global_private_data *) prv_data;
+	struct ctapi_global_private_data *priv = (struct ctapi_global_private_data *) ctx->reader_drv_data;
 
 	if (priv) {
 		int i;
@@ -622,3 +566,4 @@ struct sc_reader_driver * sc_get_ctapi_driver(void)
 	
 	return &ctapi_drv;
 }
+#endif
