@@ -19,19 +19,19 @@
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
-#ifdef HAVE_CONFIG_H
-#include <config.h>
-#endif
+#include "config.h"
+
 #include <sys/types.h>
 #include <stdlib.h>
 #include <string.h>
 #include <assert.h>
 #include <stdarg.h>
-#include <opensc/opensc.h>
-#include <opensc/cardctl.h>
-#include <opensc/log.h>
-#include <opensc/cards.h>
-#include <opensc/asn1.h>
+
+#include "libopensc/opensc.h"
+#include "libopensc/cardctl.h"
+#include "libopensc/log.h"
+#include "libopensc/cards.h"
+#include "libopensc/asn1.h"
 #include "pkcs15-init.h"
 #include "profile.h"
 
@@ -54,7 +54,7 @@ static int	cardos_store_pin(sc_profile_t *profile, sc_card_t *card,
 			const u8 *pin, size_t pin_len);
 static int	cardos_create_sec_env(sc_profile_t *, sc_card_t *,
 			unsigned int, unsigned int);
-static int	cardos_put_key(struct sc_profile *, sc_card_t *,
+static int	cardos_put_key(struct sc_profile *, sc_pkcs15_card_t *,
 			int, sc_pkcs15_prkey_info_t *,
 		       	struct sc_pkcs15_prkey_rsa *);
 static int	cardos_key_algorithm(unsigned int, size_t, int *);
@@ -117,30 +117,30 @@ tlv_len(struct tlv *tlv)
  * it's close enough to be useful.
  */
 static int
-cardos_erase(struct sc_profile *profile, sc_card_t *card)
+cardos_erase(struct sc_profile *profile, sc_pkcs15_card_t *p15card)
 {
-	return sc_pkcs15init_erase_card_recursively(card, profile, -1);
+	return sc_pkcs15init_erase_card_recursively(p15card, profile);
 }
 
 /*
  * Create the Application DF
  */
 static int
-cardos_create_dir(sc_profile_t *profile, sc_card_t *card, sc_file_t *df)
+cardos_create_dir(sc_profile_t *profile, sc_pkcs15_card_t *p15card, sc_file_t *df)
 {
 	int	r;
 
 	/* Create the application DF */
-	if ((r = sc_pkcs15init_create_file(profile, card, df)) < 0)
+	if ((r = sc_pkcs15init_create_file(profile, p15card, df)) < 0)
 		return r;
 
-	if ((r = sc_select_file(card, &df->path, NULL)) < 0)
+	if ((r = sc_select_file(p15card->card, &df->path, NULL)) < 0)
 		return r;
 
 	/* Create a default security environment for this DF.
 	 * This SE autometically becomes the current SE when the
 	 * DF is selected. */
-	if ((r = cardos_create_sec_env(profile, card, 0x01, 0x00)) < 0)
+	if ((r = cardos_create_sec_env(profile, p15card->card, 0x01, 0x00)) < 0)
 		return r;
 
 	return 0;
@@ -151,7 +151,7 @@ cardos_create_dir(sc_profile_t *profile, sc_card_t *card, sc_file_t *df)
  * See if it's good, and if it isn't, propose something better
  */
 static int
-cardos_select_pin_reference(sc_profile_t *profile, sc_card_t *card,
+cardos_select_pin_reference(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 		sc_pkcs15_pin_info_t *pin_info)
 {
 	int	preferred, current;
@@ -180,12 +180,13 @@ cardos_select_pin_reference(sc_profile_t *profile, sc_card_t *card,
  * Store a PIN
  */
 static int
-cardos_create_pin(sc_profile_t *profile, sc_card_t *card, sc_file_t *df,
+cardos_create_pin(sc_profile_t *profile, sc_pkcs15_card_t *p15card, sc_file_t *df,
 		sc_pkcs15_object_t *pin_obj,
 		const u8 *pin, size_t pin_len,
 		const u8 *puk, size_t puk_len)
 {
 	sc_pkcs15_pin_info_t *pin_info = (sc_pkcs15_pin_info_t *) pin_obj->data;
+	struct sc_card *card = p15card->card;
 	unsigned int	puk_id = CARDOS_AC_NEVER;
 	int		r;
 
@@ -219,17 +220,13 @@ cardos_create_pin(sc_profile_t *profile, sc_card_t *card, sc_file_t *df,
  * Select a key reference
  */
 static int
-cardos_select_key_reference(sc_profile_t *profile, sc_card_t *card,
+cardos_select_key_reference(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 			sc_pkcs15_prkey_info_t *key_info)
 {
-	struct sc_file	*df = profile->df_info->file;
-
 	if (key_info->key_reference < CARDOS_KEY_ID_MIN)
 		key_info->key_reference = CARDOS_KEY_ID_MIN;
 	if (key_info->key_reference > CARDOS_KEY_ID_MAX)
 		return SC_ERROR_TOO_MANY_OBJECTS;
-
-	key_info->path = df->path;
 	return 0;
 }
 
@@ -238,7 +235,7 @@ cardos_select_key_reference(sc_profile_t *profile, sc_card_t *card,
  * This is a no-op.
  */
 static int
-cardos_create_key(sc_profile_t *profile, sc_card_t *card,
+cardos_create_key(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 			sc_pkcs15_object_t *obj)
 {
 	return 0;
@@ -248,25 +245,40 @@ cardos_create_key(sc_profile_t *profile, sc_card_t *card,
  * Store a private key object.
  */
 static int
-cardos_store_key(sc_profile_t *profile, sc_card_t *card,
+cardos_store_key(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 			sc_pkcs15_object_t *obj,
 			sc_pkcs15_prkey_t *key)
 {
+	struct sc_context *ctx = p15card->card->ctx;
 	sc_pkcs15_prkey_info_t *key_info = (sc_pkcs15_prkey_info_t *) obj->data;
+	struct sc_file *file = NULL;
 	int		algorithm = 0, r;
 
 	if (obj->type != SC_PKCS15_TYPE_PRKEY_RSA) {
-		sc_error(card->ctx, "CardOS supports RSA keys only.");
+		sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "CardOS supports RSA keys only.");
 		return SC_ERROR_NOT_SUPPORTED;
 	}
 
 	if (cardos_key_algorithm(key_info->usage, key_info->modulus_length, &algorithm) < 0) {
-		sc_error(card->ctx, "CardOS does not support keys "
+		sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "CardOS does not support keys "
 			       "that can both sign _and_ decrypt.");
 		return SC_ERROR_NOT_SUPPORTED;
 	}
 
-	r = cardos_put_key(profile, card, algorithm, key_info, &key->u.rsa);
+	r = sc_select_file(p15card->card, &key_info->path, &file);
+	if (r)   {
+		sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "Failed to store key: cannot select parent DF");
+		return r;
+	}
+
+	r = sc_pkcs15init_authenticate(profile, p15card, file, SC_AC_OP_UPDATE);
+	sc_file_free(file);
+	if (r)   {
+		sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "Failed to store key: 'UPDATE' authentication failed");
+		return r;
+	}
+
+	r = cardos_put_key(profile, p15card, algorithm, key_info, &key->u.rsa);
 
 	return r;
 }
@@ -297,11 +309,12 @@ static void init_key_object(struct sc_pkcs15_prkey_rsa *key,
  * Key generation
  */
 static int
-cardos_generate_key(sc_profile_t *profile, sc_card_t *card,
+cardos_generate_key(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 		sc_pkcs15_object_t *obj,
 		sc_pkcs15_pubkey_t *pubkey)
 {
-	sc_pkcs15_prkey_info_t *key_info = (sc_pkcs15_prkey_info_t *) obj->data;
+	struct sc_context *ctx = p15card->card->ctx;
+	struct sc_pkcs15_prkey_info *key_info = (sc_pkcs15_prkey_info_t *) obj->data;
 	struct sc_pkcs15_prkey_rsa key_obj;
 	struct sc_cardctl_cardos_genkey_info args;
 	struct sc_file	*temp;
@@ -313,10 +326,10 @@ cardos_generate_key(sc_profile_t *profile, sc_card_t *card,
 	if (obj->type != SC_PKCS15_TYPE_PRKEY_RSA)
 		return SC_ERROR_NOT_SUPPORTED;
 
-	rsa_max_size = (card->caps & SC_CARD_CAP_RSA_2048) ? 2048 : 1024;
+	rsa_max_size = (p15card->card->caps & SC_CARD_CAP_RSA_2048) ? 2048 : 1024;
 	keybits = key_info->modulus_length & ~7UL;
 	if (keybits > rsa_max_size) {
-		sc_error(card->ctx, "Unable to generate key, max size is %lu",
+		sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "Unable to generate key, max size is %lu",
 			(unsigned long) rsa_max_size);
 		return SC_ERROR_INVALID_ARGUMENTS;
 	}
@@ -325,19 +338,21 @@ cardos_generate_key(sc_profile_t *profile, sc_card_t *card,
 		use_ext_rsa = 1;
 
 	if (cardos_key_algorithm(key_info->usage, keybits, &algorithm) < 0) {
-		sc_error(card->ctx, "CardOS does not support keys "
+		sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "CardOS does not support keys "
 			       "that can both sign _and_ decrypt.");
 		return SC_ERROR_NOT_SUPPORTED;
 	}
 
 	if (sc_profile_get_file(profile, "tempfile", &temp) < 0) {
-		sc_error(card->ctx, "Profile doesn't define temporary file "
+		sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "Profile doesn't define temporary file "
 				"for key generation.");
 		return SC_ERROR_NOT_SUPPORTED;
 	}
-       pin_id = sc_keycache_find_named_pin(&key_info->path, SC_PKCS15INIT_USER_PIN);
-       if (pin_id > 0) {
-		r = sc_pkcs15init_verify_key(profile, card, NULL, SC_AC_CHV, pin_id);
+
+	pin_id = sc_pkcs15init_get_pin_reference(p15card, profile, 
+			SC_AC_SYMBOLIC, SC_PKCS15INIT_USER_PIN);
+	if (pin_id >= 0) {
+		r = sc_pkcs15init_verify_secret(profile, p15card, NULL, SC_AC_CHV, pin_id);
 		if (r < 0)
 			return r;
 	}
@@ -346,13 +361,13 @@ cardos_generate_key(sc_profile_t *profile, sc_card_t *card,
 	else
 		temp->ef_structure = SC_FILE_EF_TRANSPARENT;
 
-	if ((r = sc_pkcs15init_create_file(profile, card, temp)) < 0)
+	if ((r = sc_pkcs15init_create_file(profile, p15card, temp)) < 0)
 		goto out;
 	delete_it = 1;
 
 	init_key_object(&key_obj, abignum, keybits >> 3);
 
-	r = cardos_put_key(profile, card, algorithm, key_info, &key_obj);
+	r = cardos_put_key(profile, p15card, algorithm, key_info, &key_obj);
 	if (r < 0)
 		goto out;
 
@@ -360,14 +375,14 @@ cardos_generate_key(sc_profile_t *profile, sc_card_t *card,
 	args.key_id = key_info->key_reference;
 	args.key_bits = keybits;
 	args.fid = temp->id;
-	r = sc_card_ctl(card, SC_CARDCTL_CARDOS_GENERATE_KEY, &args);
+	r = sc_card_ctl(p15card->card, SC_CARDCTL_CARDOS_GENERATE_KEY, &args);
 	if (r < 0)
 		goto out;
 
-	r = cardos_extract_pubkey(card, pubkey, temp, use_ext_rsa);
+	r = cardos_extract_pubkey(p15card->card, pubkey, temp, use_ext_rsa);
 out:
 	if (delete_it != 0)
-		sc_pkcs15init_rmdir(card, profile, temp);
+		sc_pkcs15init_rmdir(p15card, profile, temp);
 	sc_file_free(temp);
 
 	if (r < 0) {
@@ -399,7 +414,7 @@ cardos_store_pin(sc_profile_t *profile, sc_card_t *card,
 	 * "no padding required". */
 	maxlen = MIN(profile->pin_maxlen, sizeof(pinpadded));
 	if (pin_len > maxlen) {
-		sc_error(card->ctx, "invalid pin length: %u (max %u)\n",
+		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "invalid pin length: %u (max %u)\n",
 		         pin_len, maxlen);
 		return SC_ERROR_INVALID_ARGUMENTS;
 	}
@@ -516,23 +531,22 @@ cardos_create_sec_env(struct sc_profile *profile, sc_card_t *card,
 
 static int cardos_key_algorithm(unsigned int usage, size_t keylen, int *algop)
 {
-	int	sign = 0, decipher = 0;
-
-	if (usage & USAGE_ANY_SIGN) {
-		if (keylen <= 1024)
-			*algop = CARDOS_ALGO_RSA_PURE_SIG;
-		else
-			*algop = CARDOS_ALGO_EXT_RSA_SIG_PURE;
-		sign = 1;
-	}
+	/* if it is sign and decipher, we use decipher and emulate sign */
 	if (usage & USAGE_ANY_DECIPHER) {
 		if (keylen <= 1024)
 			*algop = CARDOS_ALGO_RSA_PURE;
 		else
 			*algop = CARDOS_ALGO_EXT_RSA_PURE;
-		decipher = 1;
+		return 0;
 	}
-	return (sign == decipher)? -1 : 0;
+	if (usage & USAGE_ANY_SIGN) {
+		if (keylen <= 1024)
+			*algop = CARDOS_ALGO_RSA_PURE_SIG;
+		else
+			*algop = CARDOS_ALGO_EXT_RSA_SIG_PURE;
+		return 0;
+	}
+	return -1;
 }
 
 /*
@@ -617,20 +631,24 @@ cardos_store_key_component(sc_card_t *card,
 	return sc_card_ctl(card, SC_CARDCTL_CARDOS_PUT_DATA_OCI, &args);
 }
 
-static int cardos_put_key(sc_profile_t *profile, sc_card_t *card,
+
+static int 
+cardos_put_key(sc_profile_t *profile, struct sc_pkcs15_card *p15card,
 	int algorithm, sc_pkcs15_prkey_info_t *key_info,
 	struct sc_pkcs15_prkey_rsa *key)
 {
+	struct sc_card *card = p15card->card;
 	int	r, key_id, pin_id;
 
-	key_id = key_info->key_reference;
-	pin_id = sc_keycache_find_named_pin(&key_info->path, SC_PKCS15INIT_USER_PIN);
+	pin_id = sc_pkcs15init_get_pin_reference(p15card, profile, SC_AC_SYMBOLIC, 
+			SC_PKCS15INIT_USER_PIN);
 	if (pin_id < 0)
 		pin_id = 0;
 
+	key_id = key_info->key_reference;
 	if (key_info->modulus_length > 1024 && (card->type == SC_CARD_TYPE_CARDOS_M4_2 ||
 	    card->type == SC_CARD_TYPE_CARDOS_M4_3 ||card->type == SC_CARD_TYPE_CARDOS_M4_2B ||
-	    card->type == SC_CARD_TYPE_CARDOS_M4_2C )) {
+	    card->type == SC_CARD_TYPE_CARDOS_M4_2C ||card->type == SC_CARD_TYPE_CARDOS_M4_4)) {
 		r = cardos_store_key_component(card, algorithm, key_id, pin_id, 0,
 			key->p.data, key->p.len, 0, 0);
 		if (r != SC_SUCCESS)
@@ -675,13 +693,13 @@ static int parse_ext_pubkey_file(sc_card_t *card, const u8 *data, size_t len,
 		return SC_ERROR_INVALID_ARGUMENTS;
 	data = sc_asn1_find_tag(card->ctx, data, len, 0x7f49, &ilen);
 	if (data == NULL) {
-		sc_error(card->ctx, "invalid public key data: missing tag");
+		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "invalid public key data: missing tag");
 		return SC_ERROR_INTERNAL;
 	}
 
 	p = sc_asn1_find_tag(card->ctx, data, ilen, 0x81, &tlen);
 	if (p == NULL) {
-		sc_error(card->ctx, "invalid public key data: missing modulus");
+		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "invalid public key data: missing modulus");
 		return SC_ERROR_INTERNAL;
 	}
 	pubkey->u.rsa.modulus.len  = tlen;
@@ -692,7 +710,7 @@ static int parse_ext_pubkey_file(sc_card_t *card, const u8 *data, size_t len,
 
 	p = sc_asn1_find_tag(card->ctx, data, ilen, 0x82, &tlen);
 	if (p == NULL) {
-		sc_error(card->ctx, "invalid public key data: missing exponent");
+		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "invalid public key data: missing exponent");
 		return SC_ERROR_INTERNAL;
 	}
 	pubkey->u.rsa.exponent.len  = tlen;
@@ -719,7 +737,7 @@ do_cardos_extract_pubkey(sc_card_t *card, int nr, u8 tag,
 	    || buf[2] != count + 1 || buf[3] != 0)
 		return SC_ERROR_INTERNAL;
 	bn->len = count;
-	bn->data = (u8 *) malloc(count);
+	bn->data = malloc(count);
 	if (bn->data == NULL)
 		return SC_ERROR_OUT_OF_MEMORY;
 	memcpy(bn->data, buf + 4, count);
@@ -772,8 +790,8 @@ static struct sc_pkcs15init_operations sc_pkcs15init_cardos_operations = {
 	cardos_generate_key,
 	NULL, NULL, 			/* encode private/public key */
 	NULL,				/* finalize_card */
-	NULL, NULL, NULL, NULL, NULL,	/* old style api */
-	NULL 				/* delete_object */
+	NULL, 				/* delete_object */
+	NULL, NULL, NULL, NULL, NULL  /* pkcs15init emulation */
 };
 
 struct sc_pkcs15init_operations *
