@@ -19,17 +19,17 @@
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
-#ifdef HAVE_CONFIG_H
-#include <config.h>
-#endif
+#include "config.h"
+
 #include <sys/types.h>
 #include <stdlib.h>
 #include <string.h>
 #include <assert.h>
 #include <stdarg.h>
-#include <opensc/opensc.h>
-#include <opensc/cardctl.h>
-#include <opensc/log.h>
+
+#include "libopensc/opensc.h"
+#include "libopensc/cardctl.h"
+#include "libopensc/log.h"
 #include "pkcs15-init.h"
 #include "profile.h"
 
@@ -60,7 +60,7 @@ static int	incrypto34_store_pin(sc_profile_t *profile, sc_card_t *card,
 			const u8 *pin, size_t pin_len);
 static int	incrypto34_create_sec_env(sc_profile_t *, sc_card_t *,
 			unsigned int, unsigned int);
-static int	incrypto34_put_key(struct sc_profile *, struct sc_card *,
+static int	incrypto34_put_key(struct sc_profile *, struct sc_pkcs15_card *,
 			int, sc_pkcs15_prkey_info_t *,
 		       	struct sc_pkcs15_prkey_rsa *);
 static int	incrypto34_key_algorithm(unsigned int, int *);
@@ -121,46 +121,46 @@ tlv_len(struct tlv *tlv)
  * it's close enough to be useful.
  */
 static int
-incrypto34_erase(struct sc_profile *profile, sc_card_t *card)
+incrypto34_erase(struct sc_profile *profile, sc_pkcs15_card_t *p15card)
 {
 	int r;
 	struct sc_file *file;
 	struct sc_path path;
 	memset(&file, 0, sizeof(file));
 	sc_format_path("3F00", &path);
-	if ((r = sc_select_file(card, &path, &file)) < 0)
+	if ((r = sc_select_file(p15card->card, &path, &file)) < 0)
 		return r;
-	if ((r = sc_pkcs15init_authenticate(profile, card, file, SC_AC_OP_DELETE)) < 0)
-		return sc_pkcs15init_erase_card_recursively(card, profile, -1);
+	if ((r = sc_pkcs15init_authenticate(profile, p15card, file, SC_AC_OP_DELETE)) < 0)
+		return sc_pkcs15init_erase_card_recursively(p15card, profile);
 	else
-		return sc_card_ctl(card, SC_CARDCTL_INCRYPTO34_ERASE_FILES, NULL);
+		return sc_card_ctl(p15card->card, SC_CARDCTL_INCRYPTO34_ERASE_FILES, NULL);
 }
 
 /*
  * Create the Application DF
  */
 static int
-incrypto34_create_dir(sc_profile_t *profile, sc_card_t *card, sc_file_t *df)
+incrypto34_create_dir(sc_profile_t *profile, sc_pkcs15_card_t *p15card, sc_file_t *df)
 {
 	int	r;
 	struct sc_file *file;
 	struct sc_path path;
 	memset(&file, 0, sizeof(file));
 	sc_format_path("3F00", &path);
-	if ((r = sc_select_file(card, &path, &file)) < 0)
+	if ((r = sc_select_file(p15card->card, &path, &file)) < 0)
 		return r;
-	if ((r = sc_pkcs15init_authenticate(profile, card, file, SC_AC_OP_CREATE)) < 0)
+	if ((r = sc_pkcs15init_authenticate(profile, p15card, file, SC_AC_OP_CREATE)) < 0)
 		return r;
 	/* Create the application DF */
-	if ((r = sc_pkcs15init_create_file(profile, card, df)) < 0)
+	if ((r = sc_pkcs15init_create_file(profile, p15card, df)) < 0)
 		return r;
 
-	if ((r = sc_select_file(card, &df->path, NULL)) < 0)
+	if ((r = sc_select_file(p15card->card, &df->path, NULL)) < 0)
 		return r;
 
 	/* Create a security environment for this DF.
 	*/
-	if ((r = incrypto34_create_sec_env(profile, card, 0x01, 0x00)) < 0)
+	if ((r = incrypto34_create_sec_env(profile, p15card->card, 0x01, 0x00)) < 0)
 		return r;
 
 	return 0;
@@ -171,7 +171,7 @@ incrypto34_create_dir(sc_profile_t *profile, sc_card_t *card, sc_file_t *df)
  * See if it's good, and if it isn't, propose something better
  */
 static int
-incrypto34_select_pin_reference(sc_profile_t *profile, sc_card_t *card,
+incrypto34_select_pin_reference(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 		sc_pkcs15_pin_info_t *pin_info)
 {
 	int	preferred, current;
@@ -200,8 +200,8 @@ incrypto34_select_pin_reference(sc_profile_t *profile, sc_card_t *card,
  * Store a PIN
  */
 static int
-incrypto34_create_pin(sc_profile_t *profile, sc_card_t *card, sc_file_t *df,
-		sc_pkcs15_object_t *pin_obj,
+incrypto34_create_pin(sc_profile_t *profile, sc_pkcs15_card_t *p15card, 
+		sc_file_t *df, sc_pkcs15_object_t *pin_obj,
 		const u8 *pin, size_t pin_len,
 		const u8 *puk, size_t puk_len)
 {
@@ -212,7 +212,7 @@ incrypto34_create_pin(sc_profile_t *profile, sc_card_t *card, sc_file_t *df,
 	if (!pin || !pin_len)
 		return SC_ERROR_INVALID_ARGUMENTS;
 
-	r = sc_select_file(card, &df->path, NULL);
+	r = sc_select_file(p15card->card, &df->path, NULL);
 	if (r < 0)
 		return r;
 
@@ -222,13 +222,13 @@ incrypto34_create_pin(sc_profile_t *profile, sc_card_t *card, sc_file_t *df,
 		sc_profile_get_pin_info(profile,
 				SC_PKCS15INIT_USER_PUK, &puk_info);
 		puk_info.reference = puk_id = pin_info->reference + 1;
-		r = incrypto34_store_pin(profile, card,
+		r = incrypto34_store_pin(profile, p15card->card,
 				&puk_info, INCRYPTO34_AC_NEVER,
 				puk, puk_len);
 	}
 
 	if (r >= 0) {
-		r = incrypto34_store_pin(profile, card,
+		r = incrypto34_store_pin(profile, p15card->card,
 				pin_info, puk_id,
 				pin, pin_len);
 	}
@@ -240,17 +240,13 @@ incrypto34_create_pin(sc_profile_t *profile, sc_card_t *card, sc_file_t *df,
  * Select a key reference
  */
 static int
-incrypto34_select_key_reference(sc_profile_t *profile, sc_card_t *card,
+incrypto34_select_key_reference(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 			sc_pkcs15_prkey_info_t *key_info)
 {
-	struct sc_file	*df = profile->df_info->file;
-
 	if (key_info->key_reference < INCRYPTO34_KEY_ID_MIN)
 		key_info->key_reference = INCRYPTO34_KEY_ID_MIN;
 	if (key_info->key_reference > INCRYPTO34_KEY_ID_MAX)
 		return SC_ERROR_TOO_MANY_OBJECTS;
-
-	key_info->path = df->path;
 	return 0;
 }
 
@@ -259,7 +255,7 @@ incrypto34_select_key_reference(sc_profile_t *profile, sc_card_t *card,
  * This is a no-op.
  */
 static int
-incrypto34_create_key(sc_profile_t *profile, sc_card_t *card,
+incrypto34_create_key(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 			sc_pkcs15_object_t *obj)
 {
 	return 0;
@@ -269,25 +265,26 @@ incrypto34_create_key(sc_profile_t *profile, sc_card_t *card,
  * Store a private key object.
  */
 static int
-incrypto34_store_key(sc_profile_t *profile, sc_card_t *card,
+incrypto34_store_key(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 			sc_pkcs15_object_t *obj,
 			sc_pkcs15_prkey_t *key)
 {
 	sc_pkcs15_prkey_info_t *key_info = (sc_pkcs15_prkey_info_t *) obj->data;
+	sc_card_t *card = p15card->card;
 	int		algorithm, r;
 
 	if (obj->type != SC_PKCS15_TYPE_PRKEY_RSA) {
-		sc_error(card->ctx, "Incrypto34 supports RSA keys only.");
+		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "Incrypto34 supports RSA keys only.");
 		return SC_ERROR_NOT_SUPPORTED;
 	}
 
 	if (incrypto34_key_algorithm(key_info->usage, &algorithm) < 0) {
-		sc_error(card->ctx, "Incrypto34 does not support keys "
+		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "Incrypto34 does not support keys "
 			       "that can both sign _and_ decrypt.");
 		return SC_ERROR_NOT_SUPPORTED;
 	}
 
-	r = incrypto34_put_key(profile, card, algorithm, key_info, &key->u.rsa);
+	r = incrypto34_put_key(profile, p15card, algorithm, key_info, &key->u.rsa);
 
 	return r;
 }
@@ -296,11 +293,12 @@ incrypto34_store_key(sc_profile_t *profile, sc_card_t *card,
  * Key generation
  */
 static int
-incrypto34_generate_key(sc_profile_t *profile, sc_card_t *card,
+incrypto34_generate_key(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 		sc_pkcs15_object_t *obj,
 		sc_pkcs15_pubkey_t *pubkey)
 {
 	sc_pkcs15_prkey_info_t *key_info = (sc_pkcs15_prkey_info_t *) obj->data;
+	sc_card_t *card = p15card->card;
 	struct sc_pkcs15_prkey_rsa key_obj;
 	struct sc_cardctl_incrypto34_genkey_info args;
 	struct sc_file	*temp;
@@ -309,31 +307,31 @@ incrypto34_generate_key(sc_profile_t *profile, sc_card_t *card,
 	int		algorithm, r, delete_it = 0;
 
 	if (obj->type != SC_PKCS15_TYPE_PRKEY_RSA) {
-		sc_error(card->ctx, "Incrypto34 supports only RSA keys.");
+		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "Incrypto34 supports only RSA keys.");
 		return SC_ERROR_NOT_SUPPORTED;
 	}
 
 	if (incrypto34_key_algorithm(key_info->usage, &algorithm) < 0) {
-		sc_error(card->ctx, "Incrypto34 does not support keys "
+		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "Incrypto34 does not support keys "
 			       "that can both sign _and_ decrypt.");
 		return SC_ERROR_NOT_SUPPORTED;
 	}
 
 	keybits = key_info->modulus_length & ~7UL;
 	if (keybits > RSAKEY_MAX_BITS) {
-		sc_error(card->ctx, "Unable to generate key, max size is %d",
+		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "Unable to generate key, max size is %d",
 				RSAKEY_MAX_BITS);
 		return SC_ERROR_INVALID_ARGUMENTS;
 	}
 
 	if (sc_profile_get_file(profile, "tempfile", &temp) < 0) {
-		sc_error(card->ctx, "Profile doesn't define temporary file "
+		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "Profile doesn't define temporary file "
 				"for key generation.");
 		return SC_ERROR_NOT_SUPPORTED;
 	}
 	memset(pubkey, 0, sizeof(*pubkey));
 
-	if ((r = sc_pkcs15init_create_file(profile, card, temp)) < 0)
+	if ((r = sc_pkcs15init_create_file(profile, p15card, temp)) < 0)
 		goto out;
 	delete_it = 1;
 
@@ -344,7 +342,7 @@ incrypto34_generate_key(sc_profile_t *profile, sc_card_t *card,
 	key_obj.modulus.len = keybits >> 3;
 	key_obj.d.data = abignum;
 	key_obj.d.len = keybits >> 3;
-	r = incrypto34_put_key(profile, card, algorithm, key_info, &key_obj);
+	r = incrypto34_put_key(profile, p15card, algorithm, key_info, &key_obj);
 	if (r < 0)
 		goto out;
 
@@ -368,7 +366,7 @@ incrypto34_generate_key(sc_profile_t *profile, sc_card_t *card,
 	pubkey->algorithm = SC_ALGORITHM_RSA;
 
 out:	if (delete_it) {
-		sc_pkcs15init_rmdir(card, profile, temp);
+		sc_pkcs15init_rmdir(p15card, profile, temp);
 	}
 	sc_file_free(temp);
 	if (r < 0) {
@@ -606,31 +604,32 @@ incrypto34_store_key_component(struct sc_card *card,
 }
 
 static int
-incrypto34_put_key(sc_profile_t *profile, sc_card_t *card,
+incrypto34_put_key(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 		int algorithm, sc_pkcs15_prkey_info_t *key_info,
 		struct sc_pkcs15_prkey_rsa *key)
 {
 	int	r, key_id, pin_id;
 
 	key_id = key_info->key_reference;
-	pin_id = sc_keycache_find_named_pin(&key_info->path, SC_PKCS15INIT_USER_PIN);
+	pin_id = sc_pkcs15init_get_pin_reference(p15card, profile,
+			SC_AC_SYMBOLIC, SC_PKCS15INIT_USER_PIN);
 	if (pin_id < 0)
 		pin_id = 0;
 
-	r = incrypto34_store_key_component(card, algorithm, key_id, pin_id, 0,
+	r = incrypto34_store_key_component(p15card->card, algorithm, key_id, pin_id, 0,
 			key->modulus.data, key->modulus.len, 0);
 	if (r >= 0)
 	{
-		r = incrypto34_store_key_component(card, algorithm, key_id, pin_id, 1,
+		r = incrypto34_store_key_component(p15card->card, algorithm, key_id, pin_id, 1,
 				key->d.data, key->d.len, 1);
 	}
 
 	if (SC_ERROR_FILE_ALREADY_EXISTS == r || r >=0)
 	{
-		r = incrypto34_change_key_data(card, 0x80|key_id, 0x20, key->modulus.data, key->modulus.len);
+		r = incrypto34_change_key_data(p15card->card, 0x80|key_id, 0x20, key->modulus.data, key->modulus.len);
 		if (r < 0)
 			return r;
-		r = incrypto34_change_key_data(card, 0x80|key_id, 0x21, key->d.data, key->d.len);
+		r = incrypto34_change_key_data(p15card->card, 0x80|key_id, 0x21, key->d.data, key->d.len);
 	}
 
 	return r;
@@ -655,12 +654,12 @@ incrypto34_extract_pubkey(sc_card_t *card, int nr, u8 tag,
 	 || buf[2] != count + 1 || buf[3] != 0)
 		return SC_ERROR_INTERNAL;
 	bn->len = count;
-	bn->data = (u8 *) malloc(count);
+	bn->data = malloc(count);
 	memcpy(bn->data, buf + 4, count);
 	return 0;
 }
 
-static int incrypto34_init_card(sc_profile_t *profile, sc_card_t *card)
+static int incrypto34_init_card(sc_profile_t *profile, sc_pkcs15_card_t *p15card)
 {
 	return 0;
 }
@@ -680,8 +679,8 @@ static struct sc_pkcs15init_operations sc_pkcs15init_incrypto34_operations = {
 	incrypto34_generate_key,
 	NULL, NULL, 			/* encode private/public key */
 	NULL,				/* finalize_card */
-	NULL, NULL, NULL, NULL, NULL,	/* old style api */
-	NULL 				/* delete_object */
+	NULL, 				/* delete_object */
+	NULL, NULL, NULL, NULL, NULL  /* pkcs15init emulation */
 };
 struct sc_pkcs15init_operations *
 sc_pkcs15init_get_incrypto34_ops(void)
